/*
 * @(#)DrawApplication.java 5.1
 *
 */

package CH.ifa.draw.application;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.standard.*;
import CH.ifa.draw.figures.*;
import CH.ifa.draw.util.*;

/**
 * DrawApplication defines a standard presentation for
 * standalone drawing editors. The presentation is
 * customized in subclasses.
 * The application is started as follows:
 * <pre>
 * public static void main(String[] args) {
 *     MyDrawApp window = new MyDrawApp();
 * }
 * </pre>
 */

public  class DrawApplication
        implements DrawingEditor, PaletteListener {

    private static int windowCount = 0;

    protected Frame             menuFrame, toolFrame;
    private final String        defaultDrawExtension=".draw";
    public String               drawExtension;
    private boolean             floating;
    private Vector              fDrawings = new Vector();
    private Tool                fTool;
    private Iconkit             fIconkit;
    private Menu                fDrawingsMenu;
    private TextField           fStatusLine;
    private Vector              fViewFrames = new Vector();
    private DrawingViewFrame    fViewFrame;
    private boolean             fToolButtonSticky=false;
    private ToolButton          fDefaultToolButton;
    private ToolButton          fSelectedToolButton = null;

    // the image resource path
    private static final String fgDrawPath = "/CH/ifa/draw/";
    public static final String IMAGES = fgDrawPath+"images/";

    /**
     * The index of the file menu in the menu bar.
     */
    public static final int    FILE_MENU = 0;
    /**
     * The index of the edit menu in the menu bar.
     */
    public static final int    EDIT_MENU = 1;
    /**
     * The index of the alignment menu in the menu bar.
     */
    public static final int    ALIGNMENT_MENU = 2;
    /**
     * The index of the attributes menu in the menu bar.
     */
    public static final int    ATTRIBUTES_MENU = 3;

    /**
     * The index of the drawings menu in the menu bar.
     */
    public static final int    DRAWINGS_MENU = 5;

    private String fFilename = "";

    /**
     * Constructs a drawing window with a default title.
     */
    public DrawApplication() {
       this("JHotDraw");
    }

    /**
     * Constructs a drawing window with the given title.
     */
    public DrawApplication(String title) {
       this(title, null, new String[0], null);
    }

    public DrawApplication(String title, String[] drawings) {
       this(title,null,drawings,null);
    }

    public DrawApplication(String title, String drawExtension, String[] drawings, String icon) {

        menuFrame = new Frame(title);
        fStatusLine = createStatusLine();
        menuFrame.add("South", fStatusLine);

//        toolFrame = new Frame("Tools");
//        toolFrame.setLayout(new BorderLayout());
        toolFrame = menuFrame;
        floating = toolFrame == menuFrame;
        fIconkit = new Iconkit(toolFrame);
        if (icon != null) {
            Image iconImage = fIconkit.loadImageResource(icon);
            if (iconImage == null) {
               System.err.println("Ressource icon "+icon+" could not be loaded!");
            } else {
               menuFrame.setIconImage(iconImage);
            }
        }
        Panel tools = createToolPalette();
        createTools(tools);
        toolFrame.add("Center", tools);

	MenuBar mb = new MenuBar();
	createMenus(mb);
	menuFrame.setMenuBar(mb);
        menuFrame.pack();

	// Workaround for Windows JDK pack()-bug:
	// The menu-frame is too small or too large.
        String windowResizable=System.getProperty("de.renew.windowResizable");
	if (windowResizable==null) {
            // If no parameter is given, the window is
            // resizable on Windows-Systems only.
	    if (System.getProperty("os.name").indexOf("Win")>=0) {
	       windowResizable="true";
            }
	}
        menuFrame.setResizable(Boolean.valueOf(windowResizable).booleanValue());

        //Dimension d = defaultSize();
	//setSize(d.width, d.height);
        menuFrame.show();
        if (toolFrame != menuFrame) {
           toolFrame.pack();
           toolFrame.show();
        } /* endif */

        if (drawExtension==null)
            this.drawExtension=defaultDrawExtension;
        else    
           this.drawExtension=drawExtension;
	//System.out.println("number of params: "+drawings.length);
        Drawing drawing;
        for (int i=0; i<drawings.length; ++i)
          if ((drawing = loadDrawing(drawings[i])) != null)
	     newDrawingViewFrame(drawing);
        if (fDrawings.size() == 0)
           promptNew();

	windowCount++;
        toolDone();

        addListeners();
    }

    /**
     * Registers the listeners for the main windows
     */
    protected void addListeners() {
	    menuFrame.addWindowListener(
              new WindowAdapter() {
                public void windowClosing(WindowEvent event) {
                    exit();
                }
              }
            );
            if (toolFrame != menuFrame) {
	      toolFrame.addWindowListener(
                new WindowAdapter() {
                  public void windowClosing(WindowEvent event) {
		    toolFrame.setVisible(false);
                 }
                }
              );
            } /* endif */
    }

    public Frame menuFrame() { return menuFrame; }

    protected DrawApplication newWindow() { return new DrawApplication(menuFrame.getTitle()); }

//     private Menu titleSubstitute;

//     public synchronized void setTitle(String title) {
//       menuFrame.setTitle(title);
//       titleSubstitute.setLabel(title);
//     }

    private void updateName(Drawing drawing, String filename) {
      showDrawingViewFrame(drawing);
      fFilename = noExtension(filename);
      String nameOnly = noPath(fFilename);
      fViewFrame.setTitle(nameOnly);
      drawing.setName(nameOnly);
      drawing.setFilename(filename);
      int pos = fDrawings.indexOf(drawing);
      fDrawingsMenu.getItem(pos).setLabel(nameOnly);
    }

    /**
     * Creates the standard menus. Clients override this
     * method to add additional menus.
     */
    protected void createMenus(MenuBar mb) {
		mb.add(createFileMenu());
		mb.add(createEditMenu());
		mb.add(createAlignmentMenu());
		mb.add(createAttributesMenu());
		mb.add(createDebugMenu());
		mb.add(createDrawingsMenu());
    }

    void addDrawing(Drawing drawing) {
       if (!fDrawings.contains(drawing)) {
           fDrawings.addElement(drawing);
           CheckboxMenuItem mi = new CheckboxMenuItem(drawing.getName());
           mi.addItemListener(new DrawingMenuListener(this,drawing));
           fDrawingsMenu.add(mi);
       } /* endif */
    }

// !!!!!!!!!!! This was:
//
//     void closeViewFrame(DrawingViewFrame viewFrame) {
//       viewFrame.hide();
//       viewFrame.dispose();
//       fViewFrames.removeElement(viewFrame);
//       if (viewFrame == fViewFrame) {
// 	if (fViewFrames.size() > 0)
// 	  fViewFrame = (DrawingViewFrame)fViewFrames.elementAt(0);
// 	else
// 	  fViewFrame = null;
//       }
//     }
//
//     private void removeDrawing(Drawing drawing) {
// System.out.println("Removing drawing "+drawing+"!");
//        int pos = fDrawings.indexOf(drawing);
//        fDrawingsMenu.remove(pos);
//        fDrawings.removeElementAt(pos);
//        for (int i=fViewFrames.size()-1; i>=0; --i) {
// 	 DrawingViewFrame viewFrame = (DrawingViewFrame)fViewFrames.elementAt(i);
// 	 if (viewFrame.drawingView.drawing() == drawing) {
// 	   closeViewFrame(viewFrame);
// 	 }
//        }
//        drawing.release();
//    }

    void closeViewFrame(DrawingViewFrame viewFrame) {
      viewFrame.setVisible(false);
      viewFrame.dispose();
      fViewFrames.removeElement(viewFrame);
      if (viewFrame == fViewFrame) {
        if (fViewFrames.size() > 0) {
	  setCurrentDrawing((DrawingViewFrame)fViewFrames.elementAt(0));
          fViewFrame.show();
	} else
	  setCurrentDrawing(null);
      }

      int pos = fDrawings.indexOf(viewFrame.drawingView.drawing());
      fDrawingsMenu.remove(pos);
      fDrawings.removeElementAt(pos);

      viewFrame.drawingView.drawing().release();
    }

    protected Menu createDrawingsMenu() {
        fDrawingsMenu = new Menu("Drawings");
        return fDrawingsMenu;
    }

    protected Enumeration drawings() {
      return fDrawings.elements();
    }

    private void newDrawingViewFrame(Drawing drawing) {
        Dimension d = getDrawingViewSize();
        StandardDrawingView drawingView = createDrawingView(d.width, d.height);
        addDrawing(drawing);
        DrawingViewFrame viewFrame = new DrawingViewFrame(this, drawingView, drawing);
        setCurrentDrawing(viewFrame);
    }
	  
    /**
     * Creates the file menu. Clients override this
     * method to add additional menu items.
     */
    protected Menu createFileMenu() {
		Menu menu = new Menu("File");
		MenuItem mi = new MenuItem("New Drawing", new MenuShortcut('n'));
		mi.addActionListener(
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
		            promptNew();
		        }
		    }
		);
		menu.add(mi);

// 		mi = new MenuItem("New Frame", new MenuShortcut('f'));
// 		mi.addActionListener(
// 		    new ActionListener() {
// 		        public void actionPerformed(ActionEvent event) {
// 			  newDrawingViewFrame(fViewFrame.drawingView.drawing());
// 		        }
// 		    }
// 		);
// 		menu.add(mi);

		mi = new MenuItem("Open Drawing...", new MenuShortcut('o'));
		mi.addActionListener(
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
		            promptOpen();
		        }
		    }
		);
		menu.add(mi);

		mi = new MenuItem("Save Drawing", new MenuShortcut('s'));
		mi.addActionListener(
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
		            saveDrawing(drawing());
		        }
		    }
		);
		menu.add(mi);

		mi = new MenuItem("Save Drawing As...");
		mi.addActionListener(
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
                            Drawing drawing=drawing();
		            if (promptSaveAs(drawing))
                                saveDrawing(drawing);
		        }
		    }
		);
		menu.add(mi);

		mi = new MenuItem("Save All Drawings");
		mi.addActionListener(
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
                            Enumeration drawings=drawings();
                            while (drawings.hasMoreElements()) {
                               Drawing drawing=(Drawing)drawings.nextElement();
                               if (drawing.getFilename()==null) {
                                  showDrawingViewFrame(drawing);
                                  if (!promptSaveAs(drawing))
                                      return;
                               }
                             }
                             drawings=drawings();
                             while (drawings.hasMoreElements()) {
		                saveDrawing((Drawing)drawings.nextElement());
                            }
		        }
		    }
		);
		menu.add(mi);

    	mi = new MenuItem("Close Drawing");
       	mi.addActionListener(
	   new ActionListener() {
	      public void actionPerformed(ActionEvent event) {
                 if (fViewFrame != null) {
         	    closeViewFrame(fViewFrame);
         	 }
	      }
	   }
	);
	menu.add(mi);

//     	mi = new MenuItem("Close Frame");
//        	mi.addActionListener(
// 	   new ActionListener() {
// 	      public void actionPerformed(ActionEvent event) {
// 		fViewFrame.hide();
//                 fViewFrame.dispose();
// 	      }
// 	   }
// 	);
// 	menu.add(mi);

		menu.addSeparator();
		mi = new MenuItem("Print...", new MenuShortcut('p'));
		mi.addActionListener(
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
		            print();
		        }
		    }
		);
		menu.add(mi);
		menu.addSeparator();
		mi = new MenuItem("Exit");
		mi.addActionListener(
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
		            exit();
		        }
		    }
		);
		menu.add(mi);
		return menu;
	}

    /**
     * Creates the edit menu. Clients override this
     * method to add additional menu items.
     */
    protected Menu createEditMenu() {
		CommandMenu menu = new CommandMenu("Edit");
		menu.add(new CutCommand("Cut", this), new MenuShortcut('x'));
		menu.add(new CopyCommand("Copy", this), new MenuShortcut('c'));
		menu.add(new PasteCommand("Paste", this), new MenuShortcut('v'));
		menu.addSeparator();
		menu.add(new DuplicateCommand("Duplicate", this), new MenuShortcut('d'));
		menu.add(new DeleteCommand("Delete", this));
		menu.addSeparator();
		menu.add(new GroupCommand("Group", this));
		menu.add(new UngroupCommand("Ungroup", this));
		menu.addSeparator();
		menu.add(new SendToBackCommand("Send to Back", this));
		menu.add(new BringToFrontCommand("Bring to Front", this));
		menu.addSeparator();
		menu.add(new ToggleStickyCommand("Toggle Sticky Tools",
		  this));
		return menu;
	}

    /**
     * Creates the alignment menu. Clients override this
     * method to add additional menu items.
     */
    protected Menu createAlignmentMenu() {
		CommandMenu menu = new CommandMenu("Align");
                menu.add(new ToggleGridCommand("Toggle Snap to Grid",
		  this, new Point(5,5)));
		menu.addSeparator();
		menu.add(new AlignCommand("Lefts", this, AlignCommand.LEFTS));
		menu.add(new AlignCommand("Centers", this, AlignCommand.CENTERS));
		menu.add(new AlignCommand("Rights", this, AlignCommand.RIGHTS));
		menu.addSeparator();
		menu.add(new AlignCommand("Tops", this, AlignCommand.TOPS));
		menu.add(new AlignCommand("Middles", this, AlignCommand.MIDDLES));
		menu.add(new AlignCommand("Bottoms", this, AlignCommand.BOTTOMS));
		return menu;
	}

    /**
     * Creates the debug menu. Clients override this
     * method to add additional menu items.
     */
    protected Menu createDebugMenu() {
		Menu menu = new Menu("Debug");

		MenuItem mi = new MenuItem("Simple Update");
		mi.addActionListener(
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
		            fViewFrame.drawingView.setDisplayUpdate(new SimpleUpdateStrategy());
		        }
		    }
		);
		menu.add(mi);

		mi = new MenuItem("Buffered Update");
		mi.addActionListener(
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
		            fViewFrame.drawingView.setDisplayUpdate(new BufferedUpdateStrategy());
		        }
		    }
		);
		menu.add(mi);
		return menu;
	}

    /**
     * Creates the attributes menu and its submenus. Clients override this
     * method to add additional menu items.
     */
    protected Menu createAttributesMenu() {
        Menu menu = new Menu("Attributes");
        menu.add(createColorMenu("Fill Color", "FillColor"));
        menu.add(createColorMenu("Pen Color", "FrameColor"));
        menu.add(createArrowMenu());
		menu.addSeparator();
        menu.add(createFontMenu());
        menu.add(createFontSizeMenu());
        menu.add(createFontStyleMenu());
        menu.add(createColorMenu("Text Color", "TextColor"));
        return menu;
    }

    /**
     * Creates the color menu.
     */
    protected Menu createColorMenu(String title, String attribute) {
        CommandMenu menu = new CommandMenu(title);
        for (int i=0; i<ColorMap.size(); i++)
            menu.add(
                new ChangeAttributeCommand(
                    ColorMap.name(i),
                    attribute,
                    ColorMap.color(i),
                    this
                )
            );
        return menu;
    }

    /**
     * Creates the arrows menu.
     */
    protected Menu createArrowMenu() {
        CommandMenu menu = new CommandMenu("Arrow");
        menu.add(new ChangeAttributeCommand("none",     "ArrowMode", new Integer(PolyLineFigure.ARROW_TIP_NONE),  this));
        menu.add(new ChangeAttributeCommand("at Start", "ArrowMode", new Integer(PolyLineFigure.ARROW_TIP_START), this));
        menu.add(new ChangeAttributeCommand("at End",   "ArrowMode", new Integer(PolyLineFigure.ARROW_TIP_END),   this));
        menu.add(new ChangeAttributeCommand("at Both",  "ArrowMode", new Integer(PolyLineFigure.ARROW_TIP_BOTH),  this));
        return menu;
    }

    /**
     * Creates the fonts menus. It installs all available fonts
     * supported by the toolkit implementation.
     */
    protected Menu createFontMenu() {
        CommandMenu menu = new CommandMenu("Font");
        String fonts[] = Toolkit.getDefaultToolkit().getFontList();
        for (int i = 0; i < fonts.length; i++)
            menu.add(new ChangeAttributeCommand(fonts[i], "FontName", fonts[i],  this));
        return menu;
    }

    /**
     * Creates the font style menu with entries (Plain, Italic, Bold).
     */
    protected Menu createFontStyleMenu() {
        CommandMenu menu = new CommandMenu("Font Style");
        menu.add(new ChangeAttributeCommand("Plain", "FontStyle", new Integer(Font.PLAIN), this));
        menu.add(new ChangeAttributeCommand("Italic","FontStyle", new Integer(Font.ITALIC),this));
        menu.add(new ChangeAttributeCommand("Bold",  "FontStyle", new Integer(Font.BOLD),  this));
        return menu;
    }

    /**
     * Creates the font size menu.
     */
    protected Menu createFontSizeMenu() {
        CommandMenu menu = new CommandMenu("Font Size");
        int sizes[] = { 9, 10, 12, 14, 18, 24, 36, 48, 72 };
        for (int i = 0; i < sizes.length; i++) {
            menu.add(
                new ChangeAttributeCommand(
                    Integer.toString(sizes[i]),
                    "FontSize",
                    new Integer(sizes[i]),  this)
                );
        }
        return menu;
    }

    /**
     * Creates the tool palette.
     */
    protected Panel createToolPalette() {
        Panel palette = new Panel();
        palette.setBackground(Color.lightGray);
        palette.setLayout(new PaletteLayout(2,new Point(2,2),false));
        return palette;
    }

    /**
     * Creates the tools. By default only the selection tool is added.
     * Override this method to add additional tools.
     * Call the inherited method to include the selection tool.
     * @param palette the palette where the tools are added.
     */
    protected void createTools(Panel palette) {
        Tool tool = createSelectionTool();

        fDefaultToolButton = createToolButton(IMAGES+"SEL", "Selection Tool", tool);
        palette.add(fDefaultToolButton);
    }

    /**
     * Creates the selection tool used in this editor. Override to use
     * a custom selection tool.
     */
    protected Tool createSelectionTool() {
        return new SelectionTool(this);
    }

    /**
     * Creates a tool button with the given image, tool, and text
     */
    protected ToolButton createToolButton(String iconName, String toolName, Tool tool) {
        return new ToolButton(this, iconName, toolName, tool);
    }

    /**
     * Creates the drawing view used in this application.
     * You need to override this method to use a DrawingView
     * subclass in your application. By default a standard
     * DrawingView is returned.
     */
    protected StandardDrawingView createDrawingView(int width, int height) {
        return new StandardDrawingView(this, width, height);
    }

    private void setCurrentDrawing(DrawingViewFrame viewFrame) {
         if (fViewFrame!=null) {
            int pos = fDrawings.indexOf(drawing());
            if (pos >= 0) {
                ((CheckboxMenuItem)fDrawingsMenu.getItem(pos)).setState(false);
            }
         }
         fViewFrame = viewFrame;
         if (fViewFrame!=null) {
            int pos = fDrawings.indexOf(drawing());
            ((CheckboxMenuItem)fDrawingsMenu.getItem(pos)).setState(true);
         }
     }

    public void showDrawingViewFrame(Drawing drawing) {
      Enumeration viewFrames = fViewFrames.elements();
      while (viewFrames.hasMoreElements()) {
	  DrawingViewFrame viewFrame =
            (DrawingViewFrame)viewFrames.nextElement();
	  if (viewFrame.drawingView.drawing() == drawing) {
            setCurrentDrawing(viewFrame);
            fViewFrame.show();
	    return;
	  }
      } /* endwhile */
      // Nothing was found. Create a new frame.
      newDrawingViewFrame(drawing);
    }

    /**
     * Override to define the dimensions of the drawing view.
     */
    protected Dimension getDrawingViewSize() {
        return new Dimension(400, 600);
    }

    /**
     * Creates the drawing used in this application.
     * You need to override this method to use a Drawing
     * subclass in your application. By default a standard
     * Drawing is returned.
     */
    protected Drawing createDrawing() {
        return new StandardDrawing();
    }

    /**
     * Creates the status line.
     */
    protected TextField createStatusLine() {
        TextField field = new TextField("No Tool", 40);
        field.setEditable(false);
        return field;
    }

    /**
     * Handles a user selection in the palette.
     * @see PaletteListener
     */
    public void paletteUserSelected(PaletteButton button) {
        ToolButton toolButton = (ToolButton) button;
        setSelected(toolButton);
        setTool(toolButton.tool(), toolButton.name());
    }

    /**
     * Handles when the mouse enters or leaves a palette button.
     * @see PaletteListener
     */
    public void paletteUserOver(PaletteButton button, boolean inside) {
        ToolButton toolButton = (ToolButton) button;
        if (inside)
            showStatus(toolButton.name());
        else if (fSelectedToolButton != null)
            showStatus(fSelectedToolButton.name());
    }

    /**
     * Gets the current drawing.
     * @see DrawingEditor
     */
    public Drawing drawing() {
        return fViewFrame.drawingView.drawing();
    }

    /**
     * Gets the current tool.
     * @see DrawingEditor
     */
    public Tool tool() {
        return fTool;
    }

    /**
     * Gets the current drawing view.
     * @see DrawingEditor
     */
    public DrawingView view() {
        return fViewFrame.drawingView;
    }

    /**
     * Toggle the sticky mode.
     */
    public void toggleToolButtonSticky() {
      fToolButtonSticky=!fToolButtonSticky;
    }

    /**
     * Sets the default tool of the editor.
     * @see DrawingEditor
     */
    public void toolDone() {
      // not while initializing!
      if (windowCount > 0) {
	if (fToolButtonSticky && tool()!=null) {
	  tool().deactivate();
	  tool().activate();
	} else if (fDefaultToolButton!=null) {
	  setTool(fDefaultToolButton.tool(), fDefaultToolButton.name());
	  setSelected(fDefaultToolButton);
        }
      }
    }

    /**
     * Handles a change of the current selection. Updates all
     * menu items that are selection sensitive.
     * @see DrawingEditor
     */
    public void selectionChanged(DrawingView view) {
        MenuBar mb = menuFrame.getMenuBar();
        CommandMenu editMenu = (CommandMenu)mb.getMenu(EDIT_MENU);
        editMenu.checkEnabled();
        CommandMenu alignmentMenu = (CommandMenu)mb.getMenu(ALIGNMENT_MENU);
        alignmentMenu.checkEnabled();
    }

    /**
     * Shows a status message.
     * @see DrawingEditor
     */
    public void showStatus(String string) {
        fStatusLine.setText(string);
    }

    private void setTool(Tool t, String name) {
        if (fTool != null)
            fTool.deactivate();
        fTool = t;
        if (fTool != null) {
            fStatusLine.setText(name);
            fTool.activate();
        }
    }

    private void setSelected(ToolButton button) {
        if (fSelectedToolButton != null)
            fSelectedToolButton.reset();
        fSelectedToolButton = button;
        if (fSelectedToolButton != null)
            fSelectedToolButton.select();
    }

    /**
     * Exits the application. You should never override this method
     */
    public void exit() {
        destroy();
        toolFrame.setVisible(false);
        toolFrame.dispose();
        menuFrame.setVisible(false);      // hide the Frame
        menuFrame.dispose();   // tell windowing system to free resources
        if (--windowCount <= 0)
           System.exit(0);
    }

    /**
     * Handles additional clean up operations. Override to destroy
     * or release drawing editor resources.
     */
    protected void destroy() {
    }



    /**
     * Creates a new drawing view with a new empty drawing.
     */
    public void promptNew() {
       Drawing drawing = createDrawing();
       newDrawingViewFrame(drawing);
    }

    /**
     * Shows a file dialog and opens a drawing.
     */
    public void promptOpen() {
        FileDialog dialog = new FileDialog(menuFrame, "Open File...", FileDialog.LOAD);
        dialog.show();
        String filename = dialog.getFile();
        if (filename != null) {
            filename = stripTrailingAsterisks(filename);
            String dirname = dialog.getDirectory();
            Drawing drawing;
            if ((drawing = loadDrawing(dirname + filename)) != null) {
	      newDrawingViewFrame(drawing);
            }
        }
        dialog.dispose();
    }

    public static String noExtension(String path) {
      int end = path.lastIndexOf(".");
      if (end >= 0)
          return path.substring(0,end);
      else
          return path;
    }

    public static String noPath(String path) {
      int start = path.lastIndexOf(System.getProperty("file.separator"));
      // if (start == -1) start = 0;
      return path.substring(start+1);
    }

    public static String pathOnly(String path) {
      int end = path.lastIndexOf(System.getProperty("file.separator"));
      if (end == -1)
          return "";
      return path.substring(0,end+1);
    }

   public boolean saveDrawing(Drawing drawing) {
      String filename=drawing.getFilename();
      if (filename==null) {
         if (!promptSaveAs(drawing))
            return false;
         filename=drawing.getFilename();
      }
      saveDrawing(drawing,filename);
      return true;
   }

   public void saveDrawing(Drawing drawing, String filename) {
         if (filename.endsWith(drawExtension)) {
            saveAsStorableOutput(drawing,filename);
         } else {
            saveAsObjectOutput(drawing,filename);
         }
    }

    /**
     * Shows a file dialog for saving a drawing.
     * Returns null or a complete file name.
     * If a file name is determined, the drawing's name is set accordingly.
     */
    private boolean promptSaveAs(Drawing drawing) {
        String filename=drawing.getFilename();
        if (filename==null)
           filename=pathOnly(fFilename)+drawing.getName()+drawExtension;
        toolDone();
        filename = getSavePath("Save File...",filename);
        if (filename == null)
           return false;
        if (!filename.endsWith(drawExtension) && !filename.endsWith(".ser"))
            filename += drawExtension;
        updateName(drawing,filename);
        return true;
    }

    /**
     * Prints the drawing.
     */
    public void print() {
        fTool.deactivate();
        PrintJob printJob = menuFrame.getToolkit().getPrintJob(fViewFrame, "Print Drawing", null);

        if (printJob != null) {
            Graphics pg = printJob.getGraphics();

            if (pg != null) {
                fViewFrame.drawingView.printAll(pg);
                pg.dispose(); // flush page
            }
            printJob.end();
        }
        fTool.activate();
    }

    protected String getSavePath(String title, String defaultName) {
        String path = null;
        FileDialog dialog = new FileDialog(menuFrame, title, FileDialog.SAVE);
	if (defaultName != null)
	  dialog.setFile(defaultName);
        dialog.show();
        String filename = dialog.getFile();
        if (filename != null) {
            filename = stripTrailingAsterisks(filename);
            String dirname = dialog.getDirectory();
            path = dirname + filename;
        }
        dialog.dispose();
        return path;
    }

    private String stripTrailingAsterisks(String filename) {
        // workaround for bug on NT
        if (filename.endsWith("*.*"))
            return filename.substring(0, filename.length() - 4);
        else
            return filename;
    }

    private void saveAsStorableOutput(Drawing drawing, String file) {
        // TBD: should write a MIME header
        System.out.println("Saving "+drawing.getName()+" as "+file);
        try {
            FileOutputStream stream = new FileOutputStream(file);
            StorableOutput output = new StorableOutput(stream);
            output.writeStorable(drawing);
            output.close();
        } catch (IOException e) {
            showStatus(e.toString());
        }
    }

    private void saveAsObjectOutput(Drawing drawing, String file) {
        // TBD: should write a MIME header
        try {
            FileOutputStream stream = new FileOutputStream(file);
            ObjectOutput output = new ObjectOutputStream(stream);
            output.writeObject(drawing);
            output.close();
        } catch (IOException e) {
            showStatus(e.toString());
        }
    }

    private Drawing loadDrawing(String file) {
        System.out.println("loading drawing from "+file+"...");
        showStatus("loading drawing from "+file+"...");
        Drawing drawing = null;
//        toolDone();
        String type = guessType(file);
        if (type.equals("storable"))
            drawing = readFromStorableInput(file);
        else if (type.equals("serialized"))
            drawing = readFromObjectInput(file);
        else {
            showStatus("Unknown file type of file \""+file+"\"");
            System.out.println("Unknown file type of file \""+file+"\"");
        }
        if (drawing != null) {
	   drawing.setName(noPath(noExtension(file)));
           drawing.setFilename(file);
           fFilename = noExtension(file);
        }
        return drawing;
    }

    private Drawing readFromStorableInput(String file) {
        Drawing drawing = null;
        try {
            FileInputStream stream = new FileInputStream(file);
            StorableInput input = new StorableInput(stream);
            try {
              drawing = (Drawing)input.readStorable();
            } catch (sun.io.MalformedInputException e) {
               stream.close();
               stream = new FileInputStream(file);
               input = new StorableInput(stream,false);
               drawing = (Drawing)input.readStorable();
            }
        } catch (IOException e) {
            showStatus("Error: " + e);
            System.out.println("Error: " + e);
        }
        return drawing;
    }

    private Drawing readFromObjectInput(String file) {
        Drawing drawing = null;
        try {
            FileInputStream stream = new FileInputStream(file);
            ObjectInput input = new ObjectInputStream(stream);
            drawing = (Drawing)input.readObject();
        } catch (IOException e) {
            showStatus("Error: " + e);
            System.out.println("Error: " + e);
        } catch (ClassNotFoundException e) {
            showStatus("Class not found: " + e);
            System.out.println("Error: " + e);
        }
        return drawing;
    }

    private String guessType(String file) {
        if (file.endsWith(drawExtension))
            return "storable";
        if (file.endsWith(".ser"))
            return "serialized";
        return "unknown";
    }

    public void windowOpened(WindowEvent e) {
       DrawingViewFrame viewFrame = (DrawingViewFrame)e.getWindow();
       fViewFrames.addElement(viewFrame);
       //System.out.println("New Drawing View Window: "+viewFrame.getTitle());
    }

    public void windowActivated(WindowEvent e) {
      DrawingViewFrame viewFrame = (DrawingViewFrame)e.getWindow();
      if (viewFrame != null && viewFrame != fViewFrame) {
	//System.out.println("Current Drawing View Window: "+viewFrame.getTitle());
      }
      setCurrentDrawing(viewFrame);
    }

    public void windowDeactivated(WindowEvent e) {    }

    public void windowClosing(WindowEvent e) {
      DrawingViewFrame viewFrame = (DrawingViewFrame)e.getWindow();
      // !!!!!!!!!!! This was:
      // removeDrawing(viewFrame.drawingView.drawing());
      closeViewFrame(viewFrame);
    }

    public void windowClosed(WindowEvent e) {
    }

    public void windowIconified(WindowEvent e) {
    }

    public void windowDeiconified(WindowEvent e) {
    }

}

class DrawingViewFrame extends Frame {

  StandardDrawingView drawingView;
  ScrollPane sp;
  Dimension defSize;
  //int viewNumber;
  //static int viewCounter = 0;

  DrawingViewFrame(DrawApplication editor, StandardDrawingView drawingView,
		   Drawing drawing) {
        super();
        Image iconImage = Toolkit.getDefaultToolkit()
           .createImage(editor.menuFrame().getIconImage().getSource());
        if (iconImage != null)
            setIconImage(iconImage);
        this.drawingView = drawingView;
	//viewNumber = ++viewCounter;
        addWindowListener(editor);
        sp = new ScrollPane();
        Adjustable vadjust = sp.getVAdjustable();
        Adjustable hadjust = sp.getHAdjustable();
        hadjust.setUnitIncrement(16);
        vadjust.setUnitIncrement(16);
        sp.add(drawingView);
        add(sp);
        Point pos = editor.menuFrame.getLocationOnScreen();
        Dimension siz = editor.menuFrame.getSize();
	defSize = new Dimension(siz.width, 600);
        setBounds(pos.x, pos.y + siz.height,defSize.width,defSize.height);
	doLayout();
        sp.setScrollPosition(64,64); // OS/2 wants a default scroll position!
  	setDrawing(drawing);
        show();
  }

  void setDrawing(Drawing drawing) {
    drawingView.setDrawing(drawing);
    setTitle(drawing.getName()); //+"["+viewNumber+"]");
    
   //  FigureEnumeration k = drawing.figures();
//     Rectangle minr = null;
//     while (k.hasMoreElements()) {
//       Rectangle r = k.nextFigure().displayBox();
//       if (minr == null)
// 	minr = r;
//       else
// 	minr = minr.union(r);
//     }
    if (!isShowing())
      show(); // has to be on screen to determine ViewportSize
    if (drawing.figures().hasMoreElements() && drawing instanceof Figure) {
      Dimension frame = getSize();
      Dimension inner = sp.getViewportSize();
      Dimension border = new Dimension(frame.width-inner.width,
				       frame.height-inner.height);
      Rectangle minr = ((Figure)drawing).displayBox();
      setSize(minr.width+border.width,
	      minr.height+border.height);
      validate();
      sp.setScrollPosition(minr.x, minr.y);
    } else {
      setSize(defSize.width, defSize.height);
      sp.setScrollPosition(64,64);
    }
  }
}

class DrawingMenuListener implements ItemListener {
   private DrawApplication fEditor;
   private Drawing fDrawing;

   DrawingMenuListener(DrawApplication editor, Drawing drawing) {
      fEditor = editor;
      fDrawing = drawing;
   }

   public void itemStateChanged(ItemEvent event) {
      fEditor.showDrawingViewFrame(fDrawing);
   }
}
