package CH.ifa.draw.cpn;

import java.util.*;
import java.awt.*;
import java.io.IOException;
import CH.ifa.draw.standard.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.figures.*;
import CH.ifa.draw.util.StorableInput;
import de.renew.simulator.*;
import de.renew.shadow.*;

public class CPNDrawing extends StandardDrawing {

  static ShadowCompiler compiler=null;
  private static ShadowNetSystem netSystem;
  static private ShadowLookup shadowLookup = null;

  private static Vector drawings = new Vector();
  private ShadowNet shadowNet;
  static private Simulator simulator = null;

  static {
    String netCompiler = System.getProperty("de.renew.netCompiler");
    if (netCompiler != null) {
      
      System.err.println("Trying to switch to net compiler "+
        netCompiler+"...");
      try {
	compiler = (ShadowCompiler)Class.forName(netCompiler).newInstance();
      } catch (Exception e) {
	System.err.println("Switch failed:"+e);
      }
    }
    if (compiler == null) {
      System.err.println("Using default net formalism...");
      compiler = new de.renew.java.JavaNetCompiler();
    } else {
      System.err.println("Using net compiler "+netCompiler+"...");
    }

    netSystem=new ShadowNetSystem(compiler);
  }

  public CPNDrawing() {
    super();
    shadowNet=new ShadowNet(getName(),netSystem);
    synchronized(drawings) {
      drawings.addElement(this);
    }
  }

  static CPNDrawing getDrawing(String name) {
    if (name==null) {
      return null;
    }
    synchronized(drawings) {
      for (int i=0;i<drawings.size();i++) {
	CPNDrawing currentDrawing=(CPNDrawing)drawings.elementAt(i);
	String currentName=currentDrawing.getName();
	if (name.equals(currentName)) {
	  return currentDrawing;
	}
      }
    }
    return null;
  }

  static ShadowLookup getShadowLookup() { return shadowLookup; }


  Figure locateShadow(Object errorObject) {
    FigureEnumeration figureenum=figures();
    while (figureenum.hasMoreElements()) {
      Figure figure=figureenum.nextFigure();
      if (figure instanceof ShadowHolder) {
	Figure errorFigure=((ShadowHolder)figure).locateShadow(errorObject);
	if (errorFigure!=null) {
	  return errorFigure;
	}
      }
    }
    return null;
  }

  
  static CPNDrawing findDrawing(Object errorObject) {
    // Determine the drawing containing the errorObject.
    Enumeration drawingenum=drawings.elements();
    while (drawingenum.hasMoreElements()) {
      CPNDrawing drawing=(CPNDrawing)drawingenum.nextElement();
      if (drawing.locateShadow(errorObject) != null) {
	return drawing;
      }
    }
    return null;
  }

  public void release() {
    synchronized(drawings) {
      drawings.removeElement(this);
    }
    shadowNet.discard();
    shadowNet = null;
    super.release();
    //System.err.println("Net removed");
  }

  public void setName(String name) {
    super.setName(name);
    shadowNet.setName(name);
  }

  public ShadowNet getShadowNet() {
    return shadowNet;
  }

  /**
   * Start a new simulation.
   * Return true, if the creation of the initial instance
   * was successful.
   */
  public boolean initSimulation(DrawingEditor editor) throws SyntaxException {
    simulationTerminate();
    if (netSystem.hasChanged()) {
      try {
	shadowLookup = netSystem.compile();
      } catch (SyntaxException e) {
	throw e;
      }
    }
    Net net=shadowLookup.getNet(getName());
    if (net != null) {
      // Create a new concurrent simulator.
      // One might want to configure whether a sequential
      // simulator should be started. That would exclude
      // calls to nets, but it would give a nicer user
      // interface (immediate notification of deadlocked
      // nets).
      simulator = new ConcurrentSimulator(net);
      NetInstance primaryInstance=simulator.getPrimaryInstance();
      if (primaryInstance!=null) {
	editor.showDrawingViewFrame(
          CPNInstanceDrawing.getDrawing(primaryInstance));
	return true;
      }
    } else {
      simulator = null;
    }
    return false;
  }

  public static boolean isSimulationActive() {
    return simulator != null && simulator.isActive();
  }

  public static void simulationTerminate() {
    if (simulator!=null) {
      simulator.terminateRun();
      simulator = null;
    }
    de.renew.simulator.SearchQueue.reset();
  }

  public static void simulationRun() {
    if (simulator != null) {
      simulator.startRun();
    }
  }

  public static boolean simulationStep() {
    if (simulator != null) {
      return simulator.step();
    } else {
      return false;
    }
  }

  public Figure add(Figure figure) {
    figure = super.add(figure);
    if (figure instanceof ShadowHolder) {
      ShadowHolder shadow = (ShadowHolder)figure;
      shadow.buildShadow(getShadowNet());
    }
    return figure;
  }

  /**
   * Reads the contained figures from StorableInput.
   * Verifies that all arcs have a proper shadow.
   */
  public void read(StorableInput dr) throws IOException {
    super.read(dr);
    FigureEnumeration figenum = figures();
    while (figenum.hasMoreElements()) {
      Figure figure = figenum.nextFigure();
      if (figure instanceof ArcConnection) {
	ArcConnection arc=(ArcConnection)figure;
	arc.buildShadow(getShadowNet());
      }
    }
  }

  private String macaoString(String str) {
    return String.valueOf(str.length())+":"+str;
  }

  private boolean exportTextChildren(java.io.PrintWriter output,
				     ParentFigure figure, int node,
				     String inscriptionType) {
    boolean isInit = false;
    // Look for text children:
    FigureEnumeration fe = figure.children();
    while (fe.hasMoreElements()) {
      Figure child = fe.nextFigure();
      if (child instanceof CPNTextFigure) {
	CPNTextFigure textchild = (CPNTextFigure)child;
	String name = textchild.getText();
	String texttype = null;
	int textType = textchild.getType();
	if (textType == CPNTextFigure.NAME)
	  texttype = "name";
	else if (textType == CPNTextFigure.INSCRIPTION) {
	  if (name.startsWith(":new("))
	    isInit = true;
	  texttype = inscriptionType;
	  inscriptionType = null; // only 1 inscription allowed!
	}
	if (texttype != null) {
	  if (name.length() > 0) {
	    if (texttype.equals("valuation"))
	      name = "<"+name+">";
	    output.println("CT("+macaoString(texttype)+","+node
			   +","+macaoString(name)+")");
	    Point origin = textchild.getOrigin();
	    output.println("PT("+node+","+macaoString(texttype)+","
			   +origin.x+","+origin.y+")");
	  }
	}
      }
    }
    return isInit;
  }

  private void exportArc(java.io.PrintWriter output,
			 ArcConnection arc,
			 Hashtable table,
			 Integer start, Integer end,
			 int node, boolean reverse) {
    output.println("CA("+macaoString("arc")+"," + (++node) + ","
                   + start + "," + end + ")");
    table.put(arc,new Integer(node));
    int lastpoint = arc.pointCount()-1;
    for (int i=1; i < lastpoint; ++i) {
      Point p = arc.pointAt(reverse ? lastpoint-i : i);
      output.println("PI(-1,"+node+","+(p.x-3)+","+(p.y-3)+",-1)");
    }
    exportTextChildren(output,arc,node,"valuation");
    table.put(arc,new Integer(node));
  }

  public void exportMacao(java.io.PrintWriter output) {
    output.println("DB()");
    int node = 1; // node 1 is reserved for the declaration node!
    // create initially marked place as node 2:
    output.println("CN("+macaoString("place")+","+(++node)+")");
    output.println("CT("+macaoString("marking")+","+node+","+macaoString("1")+")");

    Hashtable table = new Hashtable();
    FigureEnumeration figenum = figures();
    while (figenum.hasMoreElements()) {
      Figure figure = figenum.nextFigure();
      if (figure instanceof PlaceFigure
	   || figure instanceof TransitionFigure) {
	String type, inscrType;
	if (figure instanceof PlaceFigure) {
	  type = "place";
	  inscrType = "domain";
	} else {
	  type = "transition";
	  inscrType = null; // "guard";
	}
	output.println("CN("+macaoString(type)+"," + (++node) + ")");
	Rectangle box = figure.displayBox();
	output.println("PO(-1,"+node+","+(box.x)+","+(box.y)+","
		       +(box.x+box.width)+","+(box.y+box.height)+")");
        table.put(figure,new Integer(node));
	if (exportTextChildren(output,(ParentFigure)figure,node,inscrType)) {
	  Point center = figure.center();
	  output.println("PO(2,"+center.x+","+(center.y-3*box.height)+")");
	  output.println("CA("+macaoString("arc")+","+(node+1)+",2,"+node+")");
	  ++node;
	}
      }
    }
    figenum = figures();
    while (figenum.hasMoreElements()) {
      Figure figure = figenum.nextFigure();
      if (figure instanceof ArcConnection) {
	ArcConnection arc = (ArcConnection)figure;
	Integer start = (Integer)table.get(arc.startFigure());
	Integer end = (Integer)table.get(arc.endFigure());
	boolean twoarcs = arc.getArcType() != ShadowArc.ordinary;
	boolean reverse = arc.isReverse();
	if (twoarcs || !reverse)
	  exportArc(output,arc,table,start,end,++node,false);
	if (twoarcs || reverse)
	  exportArc(output,arc,table,end,start,++node,true);
      }
    }
    output.println("FB()");
  }


  public void exportWoflan(java.io.PrintWriter output) {
    output.println("place pinit init 1;");
    int cnt = 0;
    FigureEnumeration placeenum = figures();
    while (placeenum.hasMoreElements()) {
      Figure figure = placeenum.nextFigure();
      if (figure instanceof PlaceFigure) {
	ShadowPlace place = (ShadowPlace)((ShadowHolder)figure).getShadow();
	String name = place.getName();
	if (name == null) {
	  name = "P"+(++cnt);
	  place.setName(name);
	}
	output.println("place "+name+";");
      }
    }
    cnt = 0;
    FigureEnumeration transenum = figures();
    while (transenum.hasMoreElements()) {
      Figure figure = transenum.nextFigure();
      if (figure instanceof TransitionFigure) {
	ShadowTransition trans = (ShadowTransition)((ShadowHolder)figure).getShadow();
	String name = trans.getName();
	if (name == null) {
	  name = "T"+(++cnt);
	  trans.setName(name);
	}
	Vector inarcs = new Vector();  // ...of Placenames
	Vector outarcs = new Vector(); // dito
	// test if transition is :init(...)-transition:
	Enumeration inscrenum = trans.elements();
	while (inscrenum.hasMoreElements()) {
	  Object elem = inscrenum.nextElement();
	  if (elem instanceof ShadowInscription) {
	    if (((ShadowInscription)elem).inscr.startsWith(":init(")) {
	      inarcs.addElement("pinit");
	    }
	  }
	}
	FigureEnumeration arcenum = figures();
	while (arcenum.hasMoreElements()) {
	  figure = arcenum.nextFigure();
	  if (figure instanceof ArcConnection) {
	    ArcConnection ac = (ArcConnection)figure;
	    ShadowArc arc = (ShadowArc)ac.getShadow();
	    if (arc == null) {
               ac.updateInscriptions();
System.out.println("The shadow for arc "+ac+" was not properly constructed!");
	       continue;
	    }
	    if (arc.transition == trans) {
	      String placename = arc.place.getName();
	      if (arc.arcType == Arc.both || arc.arcType == Arc.test
		  || arc.arcType == Arc.in)
		inarcs.addElement(placename);
	      if (arc.arcType == Arc.both || arc.arcType == Arc.test
		   || arc.arcType == Arc.out)
		outarcs.addElement(placename);
	    }
	  }
	}
	output.println("trans "+name+placeVector(inarcs," in")+placeVector(outarcs," out")+";");
      }
    }
  }

  private String placeVector(Vector places, String title) {
    StringBuffer output = new StringBuffer("");
    if (places.size() > 0) {
      output.append(title);
      for (int i=places.size()-1; i>=0; --i)
	output.append(" ").append(places.elementAt(i));
    }
    return output.toString();
  }
}

