/*
 * @(#)TextFigure.java 5.1
 *
 */

package CH.ifa.draw.figures;

import java.util.*;
import java.awt.*;
import java.io.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.standard.*;
import CH.ifa.draw.util.*;

//!!! These imports are not needed. Maybe there is some reason?
//import CH.ifa.draw.cpn.ShadowHolder;
//import de.renew.simulator.*;

/**
 * A text figure.
 *
 * @see TextTool
 */
public  class TextFigure
        extends AttributeFigure
        implements ChildFigure, TextHolder {


    private int               fOriginX;
    private int               fOriginY;

    // cache of the TextFigure's size
    transient private boolean fSizeIsDirty = true;
    transient private int     fWidth;
    transient private int     fHeight;

    private String  fText;
    private Font    fFont = null;
    private static String fgCurrentFontName  = "Helvetica";
    private static int    fgCurrentFontSize  = 12;
    private static int    fgCurrentFontStyle = Font.PLAIN;
    private String  fCurrentFontName  = fgCurrentFontName;
    private int     fCurrentFontSize  = fgCurrentFontSize;
    private int     fCurrentFontStyle = fgCurrentFontStyle;
    private boolean fIsReadOnly;

    private ParentFigure  fParent = null;
    private OffsetLocator fLocator = null;

    /*
     * Serialization support.
     */
    private static final long serialVersionUID = 4599820785949456124L;
    private int textFigureSerializedDataVersion = 1;


    public TextFigure() {
        fOriginX = 0;
        fOriginY = 0;
        setAttribute("FillColor", ColorMap.color("None"));
        setAttribute("FrameColor", ColorMap.color("None"));
        fText = new String("");
        fSizeIsDirty = true;
    }

    public void moveBy(int x, int y) {
        willChange();
        basicMoveBy(x, y);
        if (fLocator != null)
            fLocator.moveBy(x, y);
        changed();
    }

    protected void basicMoveBy(int x, int y) {
        fOriginX += x;
        fOriginY += y;
    }

    public void basicDisplayBox(Point newOrigin, Point newCorner) {
        fOriginX = newOrigin.x;
        fOriginY = newOrigin.y;
    }

    public Point getOrigin() {
      return new Point(fOriginX, fOriginY);
    }

    public Rectangle displayBox() {
      textExtent();
      return new Rectangle(fOriginX, fOriginY, fWidth, fHeight);
    }

    public Rectangle textDisplayBox() {
        return displayBox();
    }

    /**
     * Tests whether this figure is read only.
     */
    public boolean readOnly() {
        return fIsReadOnly;
    }

    /**
     * Sets the read only status of the text figure.
     */
    public void setReadOnly(boolean isReadOnly) {
        fIsReadOnly = isReadOnly;
    }

    /**
     * Gets the font.
     */
    public Font getFont() {
      if (fFont == null)
	fFont = new Font(fCurrentFontName, fCurrentFontStyle, fCurrentFontSize);
      return fFont;
    }

    /**
     * Sets the font.
     */
    public void setFont(Font newFont) {
        willChange();
        fFont = newFont;
	fCurrentFontName = fFont.getName();
	fCurrentFontStyle = fFont.getStyle();
	fCurrentFontSize = fFont.getSize();
        markDirty();
        changed();
    }

    /**
     * Updates the location whenever the figure changes itself.
     */
    public void changed() {
        super.changed();
        updateLocation();
    }

    /**
     * A text figure understands the "FontSize", "FontStyle", and "FontName"
     * attributes.
     */
    public Object getAttribute(String name) {
        if (name.equals("FontSize"))
            return new Integer(fCurrentFontSize);
        if (name.equals("FontStyle"))
            return new Integer(fCurrentFontStyle);
        if (name.equals("FontName"))
            return fCurrentFontName;
        return super.getAttribute(name);
    }

    /**
     * A text figure understands the "FontSize", "FontStyle", and "FontName"
     * attributes.
     */
    public void setAttribute(String name, Object value) {
      if (name.equals("FontSize") || name.equals("FontStyle")
	  || name.equals("FontName") || name.equals("TextType")) {
	willChange();
        if (name.equals("FontSize")) {
            fCurrentFontSize = ((Integer)value).intValue();
        }
        else if (name.equals("FontStyle")) {
            int s = ((Integer)value).intValue();
            if (s == Font.PLAIN)
                fCurrentFontStyle = Font.PLAIN;
            else
                fCurrentFontStyle = fCurrentFontStyle ^ s;
        }
        else if (name.equals("FontName")) {
            fCurrentFontName = (String)value;
        }
	fFont = null;        markDirty();
        changed();

        markDirty();
        changed();
      } else {
	super.setAttribute(name, value);
      }
    }

    /**
     * Gets the text shown by the text figure.
     */
    public String getText() {
        return fText;
    }

    /**
     * Sets the text shown by the text figure.
     */
    public void setText(String newText) {
        if (!newText.equals(fText)) {
            willChange();
            fText = new String(newText);
            markDirty();
            changed();
        }
    }

    /**
     * Tests whether the figure accepts typing.
     */
    public boolean acceptsTyping() {
        return !fIsReadOnly;
    }

    public void draw(Graphics g) {
        super.draw(g);
        g.setFont(getFont());
        g.setColor((Color) getAttribute("TextColor"));
        FontMetrics metrics = g.getFontMetrics(getFont());
	int y = fOriginY + metrics.getAscent();
	StringTokenizer lines = new StringTokenizer(getText(),"\n");
	while (lines.hasMoreElements()) {
	  g.drawString(lines.nextToken(), fOriginX, y);
	  y += metrics.getHeight();
	}
    }

    public void drawBackground(Graphics g) {
        Rectangle r = displayBox();
        g.fillRect(r.x, r.y, r.width, r.height);
    }

    public void drawFrame(Graphics g) {
        Rectangle r = displayBox();
        g.drawRect(r.x, r.y, r.width-1, r.height-1);
    }

    private Dimension textExtent() {
      if (fSizeIsDirty) {
        FontMetrics metrics = Toolkit.getDefaultToolkit().getFontMetrics(getFont());
	StringTokenizer lines = new StringTokenizer(getText(),"\n");
	fWidth = 0;
	fHeight = 0;
	while (lines.hasMoreElements()) {
          fWidth = Math.max(fWidth,metrics.stringWidth(lines.nextToken()));
	  fHeight += metrics.getHeight();
	}
        fSizeIsDirty = false;
      }
      return new Dimension(fWidth, fHeight);
    }

    private void markDirty() {
        fSizeIsDirty = true;
    }

    /**
     * Gets the number of rows and columns to be overlaid when the figure
     * is edited.
     */
    public Dimension overlayRowsAndColumns() {
	int rows = 0, columns = 20;
	StringTokenizer lines = new StringTokenizer(getText(),"\n");
	while (lines.hasMoreElements()) {
          columns = Math.max(columns,lines.nextToken().length() + 3);
	  ++rows;
	}
	if (rows == 0)
	  rows = 1;
        return new Dimension(columns, rows);
    }

    public Vector handles() {
        Vector handles = new Vector();
        handles.addElement(new NullHandle(this, RelativeLocator.northWest()));
        handles.addElement(new NullHandle(this, RelativeLocator.northEast()));
        handles.addElement(new NullHandle(this, RelativeLocator.southEast()));
        handles.addElement(new FontSizeHandle(this, RelativeLocator.southWest()));
        return handles;
    }

    public void write(StorableOutput dw) {
        super.write(dw);
        dw.writeInt(fOriginX);
        dw.writeInt(fOriginY);
        dw.writeString(fText);
        dw.writeString(fCurrentFontName);
        dw.writeInt(fCurrentFontStyle);
        dw.writeInt(fCurrentFontSize);
        dw.writeBoolean(fIsReadOnly);
        dw.writeStorable(fParent);
        dw.writeStorable(fLocator);
    }

    public void read(StorableInput dr) throws IOException {
        super.read(dr);
        fOriginX = dr.readInt();
        fOriginY = dr.readInt();
        String text = dr.readString();
	setText(text);
	fCurrentFontName = dr.readString();
	fCurrentFontStyle = dr.readInt();
	fCurrentFontSize = dr.readInt();
        fFont = null;
        fIsReadOnly = dr.readBoolean();

        fParent = (ParentFigure)dr.readStorable();
        fLocator = (OffsetLocator)dr.readStorable();
        if (fParent != null)
            fParent.addChild(this);
	if (System.getProperty("de.renew.noGraphics") != null)
	  fSizeIsDirty = false;
    }

    private void readObject(ObjectInputStream s)
        throws ClassNotFoundException, IOException {

        s.defaultReadObject();

        if (fParent != null) {
            fParent.addChild(this);
	}
        markDirty();
    }

    public boolean canBeParent(ParentFigure figure) {
      while (figure != null) {
	if (figure == this)
	  return false;     // cyclic!!!
        if (figure instanceof ChildFigure)
	  figure = ((ChildFigure)figure).parent();
	else
	  break;
      }
      return true;
    }

    public boolean setParent(ParentFigure figure) {
        if (!canBeParent(figure))
	  return false;
        if (fParent != null) {
           fParent.removeChild(this);
        }
        fParent = figure;
	if (fParent == null) {
          fLocator = null;
        } else {
	  fParent.addChild(this);
          if (fLocator != null)
             fLocator.setBase(fParent.connectedTextLocator(this));
          else
	     fLocator = new OffsetLocator(fParent.connectedTextLocator(this));
	  updateLocation();
	}
	return true;

    }

    public ParentFigure parent() {
        return fParent;
    }

    /**
     * Updates the location relative to the connected figure.
     * The TextFigure is centered around the located point.
     */
    public void updateLocation() {
        if (fLocator != null) {
            Point p = fLocator.locate(fParent);
            p.x -= size().width/2 + fOriginX;
            p.y -= size().height/2 + fOriginY;

            if (p.x != 0 || p.y != 0) {
                willChange();
                basicMoveBy(p.x, p.y);
                changed();
            }
        }
    }

    public void release() {
        super.release();
        if (fParent != null) {
            fParent.removeChild(this);
	}
    }

    /**
     * Creates the current font to be used for new text figures.
     */
    static public Font createCurrentFont() {
        return new Font(fgCurrentFontName, fgCurrentFontStyle, fgCurrentFontSize);
    }

    /**
     * Sets the current font name
     */
    static public void setCurrentFontName(String name) {
        fgCurrentFontName = name;
    }

    /**
     * Sets the current font size.
     */
    static public void setCurrentFontSize(int size) {
        fgCurrentFontSize = size;
    }

    /**
     * Sets the current font style.
     */
    static public void setCurrentFontStyle(int style) {
        fgCurrentFontStyle = style;
    }


    public void figureChanged(FigureChangeEvent e) {
        updateLocation();
    }

    public void figureRemoved(FigureChangeEvent e) {
        if (listener() != null)
            listener().figureRequestRemove(new FigureChangeEvent(this));
    }

    public void figureRequestRemove(FigureChangeEvent e) {}
    public void figureInvalidated(FigureChangeEvent e) {}
    public void figureRequestUpdate(FigureChangeEvent e) {}


}
