package de.renew.java;

import java.util.Vector;
import java.lang.reflect.*;
import de.renew.util.Types;
import de.renew.unify.Tuple;
import de.renew.function.*;
import de.renew.simulator.*;

class JavaHelper {
  static ParseException makeParseException(String msg, Token token) {
    if (token!=null) {
      ParseException ex=new ParseException("Error in line "+
        token.next.beginLine+
        ", column "+token.next.beginColumn+":\n"+msg);
      ex.currentToken=token;
      return ex;
    } else {
      return new ParseException(msg);
    }
  }

  static TypedExpression makeTypedBinary(TypedExpression left,
    TypedExpression right, de.renew.unify.Function fun, Class type)
  {
    return new TypedExpression(type,
      new CallExpression(
	new TupleExpression(left.getExpression(),right.getExpression()),
        fun));   
  }

  static TypedExpression makeBooleanBinary(TypedExpression left,
    TypedExpression right, de.renew.unify.Function fun)
  {
    return makeTypedBinary(left,right,fun,Boolean.TYPE);
  }

  static TypedExpression makeBinary(TypedExpression left,
    TypedExpression right, de.renew.unify.Function fun,
    Token errToken) throws ParseException
  {
    Class type;
    if (!left.isTyped() || !right.isTyped()) {
      type=Types.UNTYPED;
    } else if (left.getType()==Boolean.TYPE || right.getType()==Boolean.TYPE) {
      if (left.getType()!=Boolean.TYPE || right.getType()!=Boolean.TYPE) {
	throw makeParseException("Operator types do not match.",errToken);
      }
      type=Boolean.TYPE;
    } else if (left.getType()==Double.TYPE || right.getType()==Double.TYPE) {
      type=Double.TYPE;
    } else if (left.getType()==Float.TYPE || right.getType()==Float.TYPE) {
      type=Float.TYPE;
    } else if (left.getType()==Long.TYPE || right.getType()==Long.TYPE) {
      type=Long.TYPE;
    } else {
      type=Integer.TYPE;
    }

    return makeTypedBinary(left,right,fun,type);
  }

  static Class unaryNumericPromotion(Class old, Token errToken)
    throws ParseException
  {
    // Accept untyped expressions.
    if (old==Types.UNTYPED) {
      return old;
    }
    // Exclude reference types and boolean.
    if (old==null || !old.isPrimitive()) {
      throw makeParseException("Primitive type expected.",errToken);
    }
    if (old==Boolean.TYPE) {
      throw makeParseException("Numeric type expected.",errToken);
    }
    // Convert small types to int.
    if (old==Byte.TYPE || old==Short.TYPE || old==Character.TYPE) {
      return Integer.TYPE;
    } else {
      return old;
    }
  }

  static Class unaryIntegralPromotion(Class old, Token errToken)
    throws ParseException {
    if (old==Float.TYPE || old==Double.TYPE) {
      throw makeParseException("Integral type expected.",errToken);
    }
    return unaryNumericPromotion(old,errToken);
  }

  // ensureNonVoid ensures that the given expression is not
  // of the type void.
  static void ensureNonVoid(TypedExpression expr, Token errToken)
    throws ParseException
  {
    if (expr.getType()==Void.TYPE) {
      throw makeParseException("Expression of type void not allowed here.",
        errToken);
    }
  }

  // ensureConvertability ensures that the given typed expression
  // provides a result the can be converted to the given type.
  // 
  // If the expression is untyped, the method always succeeds.
  static void ensureConvertability(Class clazz,
    TypedExpression expr, Token errToken) throws ParseException
  {
    if (expr.isTyped() &&
      !Types.allowsWideningConversion(expr.getType(),clazz))
    {
      throw makeParseException("Cannot convert "+
	makeTypeErrorString(expr.getType())+
	" to "+makeTypeErrorString(clazz)+".",errToken);
    }  
  }

  static void ensureEnumerateability(TypedExpression expr,
    Token errToken) throws ParseException
  {
    if (expr.isTyped()) {
      if (expr.getType()==null || !expr.getType().isPrimitive() ||
	expr.getType()==Double.TYPE || expr.getType()==Float.TYPE)
      {
	throw makeParseException("Enumerable type expected.",errToken);
      }
    }
  }
  
  static void ensureBinaryMatch(TypedExpression left,
    TypedExpression right, Token errToken) throws ParseException
  {
    if (left.isTyped() && right.isTyped()) {
      Class leftType=left.getType();
      Class rightType=right.getType();
      boolean leftRef=(leftType==null || !leftType.isPrimitive());
      boolean rightRef=(rightType==null || !rightType.isPrimitive());
      if (leftRef ^ rightRef) {
        throw makeParseException("Operator types do not match.",errToken);
      }
      if (leftType==Boolean.TYPE ^ rightType==Boolean.TYPE) {
        throw makeParseException("Operator types do not match.",errToken);
      }
    }
  }

  static private boolean allowsInvertibleCast(Class from, Class to) {
    // Only called from makeExplicitCastExpression.
    // Neither class may be untyped.
    if (to.isPrimitive()) {
      return Types.allowsLosslessWidening(from,to);
    } else {
      if (from==null) return true;
      return !from.isPrimitive();
    }
  }
      
  static TypedExpression makeExplicitCastExpression(Class clazz,
    TypedExpression expr, Token errToken) throws ParseException
  {
    // We must distiguish the following cases:
    //  - An identity cast. No action.
    //  - The cast does not loose information.
    //    (Typical: byte -> int, String -> Object, Object -> String)
    //  - The cast might loose information.
    //    (Typical: int -> byte, long -> double)
    //  - The cast is impossible.
    //    (Typical: int -> Object, String -> Class)
    if (expr.getType()==clazz) {
      return expr;
    } else if (!expr.isTyped()) {
      if (clazz.isPrimitive()) {
	// We do not know which class the untyped expression
	// will take at runtime, so we must restrict ourselves to
	// directed information transfer.
	return new TypedExpression(clazz,
	  new CallExpression(expr.getExpression(),
	    new CastFunction(clazz)));
      } else {
	// It is still save to allow bidirectional communication.
	// With references types, the value is only checked
	// and handed through.
	return new TypedExpression(clazz,
	  new GuardExpression(expr.getExpression(),
	    new InstanceofFunction(clazz)));
      }
    } else if (Types.allowsCast(expr.getType(),clazz)) {
      if (clazz.isPrimitive()) {
	if (Types.allowsLosslessWidening(expr.getType(),clazz)) {
	  return new TypedExpression(clazz,
	    new InvertibleExpression(expr.getExpression(),
	      new CastFunction(clazz),
	      new CastFunction(expr.getType())));
	} else {
	  return new TypedExpression(clazz,
	    new CallExpression(expr.getExpression(),
	      new CastFunction(clazz)));
	}
      } else {
	// We must guard against two cases.
	// The guard may fail and a value might be illegally
	// fed backwards into the equation.
	return new TypedExpression(clazz,
	  new GuardExpression(makeGuardedExpression(expr),
	    new InstanceofFunction(clazz)));
      }
    } else {
      throw makeParseException("Cannot cast "+expr.getType()+
	" to "+clazz+".",errToken);
    }
  }

  static TupleExpression makeTupleExpression(Vector vector) {
    Object[] result=new Object[vector.size()];
    vector.copyInto(result);
    for (int i=0;i<result.length;i++) {
      TypedExpression expr=(TypedExpression)result[i];
      result[i]=expr.getExpression();
    }
    return new TupleExpression(new Tuple(result,null));
  }
  
  static Expression makeGuardedExpression(TypedExpression expr) {
    // The typed expression is wrapped in a guard expression
    // to ensure that no illegally typed value is propagated
    // backwards into the expression.
    if (expr.isTyped() && expr.getType()!=null) {
      return new GuardExpression(expr.getExpression(),
        new InstanceofFunction(expr.getType()));
    } else {
      return expr.getExpression();
    }
  }

  static TupleExpression makeGuardedTupleExpression(Vector vector) {
    // We receive a vector of typed expressions and convert it
    // into a tuple expression that is sufficiently guarded.
    Object[] result=new Object[vector.size()];
    for (int i=0;i<result.length;i++) {
      result[i]=makeGuardedExpression((TypedExpression)vector.elementAt(i));
    }
    return new TupleExpression(new Tuple(result,null));
  }
  
  static Class[] makeTypeArray(Vector vector) {
    Class[] types=new Class[vector.size()];
    for (int i=0;i<types.length;i++) {
      TypedExpression expr=(TypedExpression)vector.elementAt(i);
      if (!expr.isTyped()) {
	// If one component is untyped, then the whole vector is untyped.
	return null;
      }
      types[i]=expr.getType();
    }
    return types;
  }

  static Class increaseArrayLevel(Class clazz, Token errToken)
    throws ParseException
  {
    try {
      return java.lang.reflect.Array.newInstance(clazz,0).getClass();
    }
    catch (Exception e) {
      throw makeParseException("Could not create class.",errToken);
    }
  }

  static String makeTypeErrorString(Class type) {
    if (type==Types.UNTYPED) {
      return "an untyped expression";
    } else {
      return "an expression of type "+type;
    }
  }

  static private TypedExpression makeTypedEquality(TypedExpression left,
    TypedExpression right, Token errToken) throws ParseException
  {
    // It is assumed that this routine is called form
    // makeEqualityAssertion where the proper checks have been
    // undertaken. Both expressions are typed, different,
    // and the right expression allows a lossless widening conversion
    // to the type of the left expression.

    Class leftType=left.getType();
    Class rightType=right.getType();

    Expression castedRight;    
    if (rightType.isPrimitive()) {
      // The equality will only hold if the two sides result in the
      // same value. The conversion that is required is done in such
      // a way that it is invertible. The two casts will automatically
      // detect cases where one side is out of range for the other side.
      castedRight=new InvertibleExpression(right.getExpression(),
	new CastFunction(leftType),
	new CastFunction(rightType));
    } else {
      // There is no conversion involved, but we must make sure
      // that no illegal value is fed back into the right expression
      // that has a narrower type.
      castedRight=makeGuardedExpression(right);
    }

    return new TypedExpression(left.getType(),
      new EqualsExpression(left.getExpression(),castedRight));
  }

  static TypedExpression makeEqualityAssertion(TypedExpression left,
    TypedExpression right, Token errToken) throws ParseException
  {
    Class leftType=left.getType();
    Class rightType=right.getType();
    if (leftType==Void.TYPE || rightType==Void.TYPE) {
      throw makeParseException("Expression of type void not allowed here.",
	errToken);
    }
    if (!left.isTyped() || !right.isTyped() ||
      Types.allowsIdentityConversion(leftType,rightType))
    {
      return new TypedExpression(leftType,
        new EqualsExpression(left.getExpression(),right.getExpression()));
    }
    // Both expressions are typed.
    if (Types.allowsLosslessWidening(rightType,leftType)) {
      return makeTypedEquality(left,right,errToken);
    }
    if (Types.allowsLosslessWidening(leftType,rightType)) {
      return makeTypedEquality(right,left,errToken);
    }
    throw makeParseException("Type mismatch in assignment.",errToken);
  }

  static TypedExpression makeSinglePartExpression(PrimaryPart firstPart,
    TypedExpression right, ParsedDeclarationNode declNode)
    throws ParseException
  {
    switch (firstPart.type) {
    case PrimaryPart.EXPR:
      if (right==null) {
	return (TypedExpression)firstPart.obj;
      } else {
	throw makeParseException("Invalid left hand side of assignment.",
	  firstPart.token);
      }
      case PrimaryPart.NAME:
        Object meaning=declNode.interpreteName((String)firstPart.obj);
        if (meaning instanceof LocalVariable) {
	  Class type=declNode.findType((LocalVariable)meaning);

          if (right==null) {
	    return new TypedExpression(type,
	      new VariableExpression((LocalVariable)meaning));
	  } else {
	    // This is called in an action inscription
	    // where we do not want to propagate information
	    // backwards due to consistency reasons.
	    //
	    // Furthermore, an inscription like "action x=y"
	    // is accepted as compile time type safe, if
	    // x is assignable from y. But backward information
	    // transfer would ruin this condition.
	    // 
	    // Note that the inscription "x=y" is interpreted
	    // and implemented differently.
	    if (type==Types.UNTYPED) {
	      return new TypedExpression(type,
		new EqualsExpression(
                  new VariableExpression((LocalVariable)meaning),
	          new CallExpression(right.getExpression(),Identity.FUN)));
	    } else if (right.isTyped() &&
                Types.allowsWideningConversion(right.getType(),type)) 
	    {
	      return new TypedExpression(type,
		new EqualsExpression(
                  new VariableExpression((LocalVariable)meaning),
	          new CallExpression(right.getExpression(),
		    new CastFunction(type))));
	    } else {
	      throw makeParseException("Cannot assign "+
                makeTypeErrorString(right.getType())+" to "+
		makeTypeErrorString(type)+".",firstPart.token);
	    }
	  }
	} else {
	  throw makeParseException("No such variable.",firstPart.token);
	}
      default:
        throw new RuntimeException("Malformed primary vector.");
      }
  }

  static TypedExpression makeExpression(Vector vector, TypedExpression right,
    ParsedDeclarationNode declNode, Token rightErrToken) throws ParseException
  {
    if (right!=null) {
      ensureNonVoid(right,rightErrToken);
    }

    PrimaryPart firstPart=(PrimaryPart)vector.elementAt(0);

    if (vector.size()==1) {
      return makeSinglePartExpression(firstPart,right,declNode);
    }

    // Ok, so it is a long vector, we must analyse it part by part.
    int m=vector.size();
    if (right!=null) {
      // We must create an assignment, so we leave a single
      // item in the queue, if we haven't read it so far.
      m--;
    }

    // How many parts did we analyse?
    int i=1;

    // During the composition the following variable is either
    // null or a class or a local variable or a typed expression.
    // Ultimately, it will hold an expression.
    Object composed=null;

    switch (firstPart.type) {
    case PrimaryPart.EXPR:
      composed=firstPart.obj;
      break;
    case PrimaryPart.NAME:
      String name=(String)firstPart.obj;
      composed=declNode.interpreteName(name);

      // Where did we find the longest match?
      int lasti=i;

      while (i<m) {
        PrimaryPart part=(PrimaryPart)vector.elementAt(i);
        i++;
        if (part.type!=PrimaryPart.NAME) {
          // Leave the while loop.
	  // I hate to do this, but here it is really the
	  // best solution.
	  break; 
        }
        name=name+"."+(String)part.obj;
        Object newComposed=declNode.interpreteName(name);
	if (newComposed!=null) {
	  composed=newComposed;
	  lasti=i;
	}
      }

      // Reset the list pointer to the last value
      // where we parsed a name.
      i=lasti;

      // At this time composed might be null if no appropriate name
      // was found.
      break;
    default:
      throw new RuntimeException("Malformed primary vector.");
    }

    if (composed==null) {
      throw makeParseException("No such class or variable.",firstPart.token);
    }

    // Let's get rid of local variables.
    if (composed instanceof LocalVariable) {
      composed=new TypedExpression(declNode.findType((LocalVariable)composed),
	new VariableExpression((LocalVariable)composed));
    }

    // From now on, the composed object can only be a class or
    // a typed expression.

    while (i<m) {
      PrimaryPart part=(PrimaryPart)vector.elementAt(i);
      i++;
      switch (part.type) {
      case PrimaryPart.EXPR:
	throw new RuntimeException("Malformed primary vector.");
      case PrimaryPart.NAME:
	// Does an argument list follow?
	Vector args=null;
	if (i<vector.size()) {
	  PrimaryPart additionalPart=(PrimaryPart)vector.elementAt(i);
	  if (additionalPart.type==PrimaryPart.CALL) {
            i++;
	    args=(Vector)additionalPart.obj;
	  }
	}

	if (args!=null) {
	  Class[] types=makeTypeArray(args);
	  // Method call.
	  if (composed instanceof Class) {
	    if (types==null) {
	      composed=new TypedExpression(Types.UNTYPED,
		new CallExpression(makeTupleExpression(args),
		  new StaticMethodFunction((String)part.obj,(Class)composed)));
	    } else {
	      Method method;
	      try {
		method=Executor.findBestMethod((Class)composed,
		  (String)part.obj,types);
	      }
	      catch (NoSuchMethodException e) {
		throw makeParseException("No such method.",part.token);
	      }
	      if ((method.getModifiers() & Modifier.STATIC)==0) {
		throw makeParseException("Cannot make static call to "+
		  "instance method.",part.token);
	      }
	      composed=new TypedExpression(method.getReturnType(),
		new CallExpression(makeTupleExpression(args),
		  new StaticMethodFunction(method)));
	    }
	  } else {
	    // composed instanceof TypedExpression
	    if (types==null || !((TypedExpression)composed).isTyped()) {
	      composed=new TypedExpression(Types.UNTYPED,
		new CallExpression(
                  new TupleExpression(
                    ((TypedExpression)composed).getExpression(),
		    makeTupleExpression(args)),
                  new MethodFunction((String)part.obj)));
	    } else {
	      Method method;
	      try {
		method=Executor.findBestMethod(
                ((TypedExpression)composed).getType(),
		(String)part.obj,types);
	      }
	      catch (NoSuchMethodException e) {
		throw makeParseException("No such method.",part.token);
	      }
	      // Even if the method is a static method,
	      // it is invoked like an ordinary instance method.
	      // This ensures the greatest compatibility to Java.
              composed=new TypedExpression(method.getReturnType(),
		new CallExpression(
                  new TupleExpression(
                    ((TypedExpression)composed).getExpression(),
		    makeTupleExpression(args)),
		  new MethodFunction(method)));
	    }
	  }
	} else {
	  // Field access.
	  if (composed instanceof Class) {
	    Class clazz=(Class)composed;
	    Field field=null;
	    try {
	      field=clazz.getField((String)part.obj);
	    }
	    catch (Exception e) {
	      throw makeParseException("No such field.",part.token);
	    }
	    composed=new TypedExpression(field.getType(),
	      new NoArgExpression(new StaticFieldFunction(field)));
	  } else {
	    // composed instanceof TypedExpression
	    if (((TypedExpression)composed).isTyped()) {
	      Class type=((TypedExpression)composed).getType();
	      Field field=null;
	      try {
		field=type.getField((String)part.obj);
	      }
	      catch (Exception e) {
		throw makeParseException("No such field.",part.token);
	      }
	      composed=new TypedExpression(field.getType(),
		new CallExpression(((TypedExpression)composed).getExpression(),
		  new FieldFunction(field)));
	    } else {
	      composed=new TypedExpression(Types.UNTYPED,
		new CallExpression(((TypedExpression)composed).getExpression(),
		  new FieldFunction((String)part.obj)));
	    }
	  }
	}
        break;
      case PrimaryPart.ARRAY:
	if (!(composed instanceof TypedExpression)) {
	  throw new RuntimeException("No such class or variable.");
	}
        Class type=Types.UNTYPED;
	if (((TypedExpression)composed).isTyped()) {
	  type=((TypedExpression)composed).getType();
	  if (!type.isArray()) {
	    throw makeParseException("Not an array.",part.token);
	  }
	  type=type.getComponentType();
	}
	ensureConvertability(Integer.TYPE,(TypedExpression)part.obj,
	  part.token);
	composed=new TypedExpression(type,
	  new CallExpression(
	    new TupleExpression(
              ((TypedExpression)composed).getExpression(),
	      ((TypedExpression)part.obj).getExpression()),
	    ArrayFunction.FUN));
	break;
      case PrimaryPart.CALL:
	throw makeParseException("Bad method call or no such method.",
	  part.token);
      default:
	throw new RuntimeException("Malformed primary vector.");
      }
    }

    if (right!=null) {
      if (i==vector.size()) {
	throw makeParseException("Invalid left hand side of assignment.",
	  rightErrToken);
      }

      // Process the very last item in the queue.
      PrimaryPart part=(PrimaryPart)vector.elementAt(i);
      switch (part.type) {
      case PrimaryPart.NAME:
	if (composed instanceof TypedExpression) {
	  if (((TypedExpression)composed).isTyped()) {
            Class type=((TypedExpression)composed).getType();
            Field field=null;
            try {
	      field=type.getField((String)part.obj);
	    }
	    catch (Exception e) {
	      throw makeParseException("No such field.",part.token);
	    }
	    ensureConvertability(field.getType(),right,rightErrToken);
	    composed=new TypedExpression(field.getType(),
	      new CallExpression(
                new TupleExpression(
		  ((TypedExpression)composed).getExpression(),
		  right.getExpression()),
                new FieldWriteFunction(field)));
	  } else {
	    composed=new TypedExpression(Types.UNTYPED,
	      new CallExpression(
                new TupleExpression(
		  ((TypedExpression)composed).getExpression(),
		  right.getExpression()),
                new FieldWriteFunction((String)part.obj)));
	  }
	} else {
	  // composed instanceof Class
	  Class clazz=(Class)composed;
	  Field field=null;
	  try {
	    field=clazz.getField((String)part.obj);
	  }
	  catch (Exception e) {
	    throw makeParseException("No such field.",part.token);
	  }
	  ensureConvertability(field.getType(),right,rightErrToken);
	  composed=new TypedExpression(field.getType(),
	    new CallExpression(
	      right.getExpression(),
	      new StaticFieldWriteFunction(field)));
	}
        break;
      case PrimaryPart.ARRAY:
	if (!(composed instanceof TypedExpression)) {
	  throw new RuntimeException("No such class or variable.");
	}
        Class type=Types.UNTYPED;
	if (((TypedExpression)composed).isTyped()) {
	  type=((TypedExpression)composed).getType();
	  if (!type.isArray()) {
	    throw makeParseException("Not an array.",part.token);
	  }
	  type=type.getComponentType();
	}
	ensureConvertability(Integer.TYPE,(TypedExpression)part.obj,
	  part.token);
	ensureConvertability(type,right,rightErrToken);
	composed=new TypedExpression(type,
	  new CallExpression(
	    new TupleExpression(
              ((TypedExpression)composed).getExpression(),
              ((TypedExpression)part.obj).getExpression(),
	      right.getExpression()),
	    ArrayWriteFunction.FUN));
        break;
      default:
	throw new RuntimeException("Malformed primary vector.");
      }
    }
    
    if (!(composed instanceof TypedExpression)) {
      throw makeParseException("No such class or variable.",firstPart.token);
    }
    
    return (TypedExpression)composed;
  }
}
