package de.renew.java;

import collections.*;
import de.renew.util.Types;
import de.renew.simulator.*;
import de.renew.shadow.*;
import de.renew.unify.*;
import de.renew.function.*;
import java.io.*;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

public class JavaNetCompiler implements ShadowCompiler {
  private ShadowLookup lookup;
  private Hashtable placeTypes; // Maps shadow places to types.
  private ShadowNetSystem netSystem;

  public JavaNetCompiler() {}
 
  protected InscriptionParser makeParser(String inscr) {
    return new JavaNetParser(new StringReader(inscr));
  }

  SyntaxException makeSyntaxException(ParseException e) {
    Token t=e.currentToken.next;
    return new SyntaxException(e.getMessage(),null,
      t.beginLine,t.beginColumn);
  }

  // Check the validity of certain inscriptions.
  public void checkDeclarationNode(String inscr) throws SyntaxException {
    parseDeclarationNode(inscr);
  }

  public void checkArcInscription(String inscr, ShadowNet shadowNet)
    throws SyntaxException
  {
    parseArcInscription(inscr,makeDeclarationNode(shadowNet));
  }

  public void checkTransitionInscription(String inscr, ShadowNet shadowNet)
    throws SyntaxException 
  {
    makeInscriptions(inscr,null,false,makeDeclarationNode(shadowNet));
  }

  public void checkPlaceInscription(String inscr, ShadowNet shadowNet)
    throws SyntaxException 
  {
    ParsedDeclarationNode declNode=makeDeclarationNode(shadowNet);
    parsePlaceInscription(inscr,declNode);
  }

  // Do the actual compilation.

  ParsedDeclarationNode compile(ShadowDeclarationNode declaration)
    throws SyntaxException
  {
    try {
      return parseDeclarationNode(declaration.inscr);
    }
    catch (SyntaxException e) {
      throw e.addObject(declaration);
    }
  }

  void compile(ShadowPlace shadowPlace, ParsedDeclarationNode declaration)
    throws SyntaxException
  {
    Net net=lookup.getNet(shadowPlace.shadowNet.getName());

    // Determine the name.
    String pname=shadowPlace.getName();
    if (pname==null) {
      pname=lookup.makePlaceName();
    }

    // Create the new place;
    Place place=new Place(net,pname);
    lookup.set(shadowPlace,place);

    // Store the initialisation inscriptions for later use.
    Vector parsedInscriptions=new Vector();
    Vector errorShadows=new Vector();

    // Insert the inscriptions.
    CollectionEnumeration enum=shadowPlace.elements();
    while (enum.hasMoreElements()) {
      Object elem=enum.nextElement();
      if (elem instanceof ShadowInscription) {
	String inscr=((ShadowInscription)elem).inscr;

	try {
	  Object typeOrExpression=parsePlaceInscription(inscr,declaration);

	  if (!(typeOrExpression instanceof Class) &&
	    !(typeOrExpression instanceof TypedExpression))
	  {
	    throw new RuntimeException(
	      "parsePlaceInscription returned a bad object: "+
		typeOrExpression);
	  }

	  parsedInscriptions.addElement(typeOrExpression);
	  errorShadows.addElement(elem);
	} catch (SyntaxException e) {
	  throw e.addObject((ShadowInscription)elem);
	}
      } else {
	throw new SyntaxException("Unsupported place inscription",
	  null).addObject(shadowPlace).addObject(elem);
      }
    }

    // Remember the type of this place.
    Class type=Types.UNTYPED;
    int typeCount=0;

    for (int i=0; i<parsedInscriptions.size(); i++) {
      if (parsedInscriptions.elementAt(i) instanceof Class) {
	type=(Class)parsedInscriptions.elementAt(i);
	typeCount++;
      }
    }

    // Check for multiple types.
    if (typeCount>1) {
      SyntaxException e=new SyntaxException("Place is typed more than once.",
	null);
      for (int i=0; i<parsedInscriptions.size(); i++) {
	if (parsedInscriptions.elementAt(i) instanceof Class) {
	  e.addObject(errorShadows.elementAt(i));
	}
      }
      throw e;
    }

    // Remember the type information for this place.
    if (type!=Types.UNTYPED) {
      placeTypes.put(shadowPlace,type);
    }

    // Now add the initial markings.
    for (int i=0; i<parsedInscriptions.size(); i++) {
      if (parsedInscriptions.elementAt(i) instanceof TypedExpression) {
	TypedExpression expr=(TypedExpression)parsedInscriptions.elementAt(i);
	
	Expression castedExpression=null;
	try {
	  castedExpression=makeCastedOutputExpression(type,
	    expr);
	} catch (SyntaxException e) {
	  throw e.addObject(errorShadows.elementAt(i));
	}
	place.add(new InitialMarkingExpression(castedExpression));
      }
    }
  }

  Class getType(ShadowPlace place) {
    if (placeTypes.containsKey(place)) {
      return (Class)placeTypes.get(place);
    } else {
      return Types.UNTYPED;
    }
  }

  void compile(ShadowTransition transition, ParsedDeclarationNode declaration,
    Net net) throws SyntaxException 
  {
    // Determine the name.
    String tname=transition.getName();
    if (tname==null) {
      tname=lookup.makeTransitionName();
    }

    // Create the new transition.
    Transition createdTransition=new Transition(net,tname);
    createdTransition.setTrace(transition.getTrace());
    lookup.set(transition,createdTransition);

    // Prepare a list to hold the inscriptions.
    Vector parsedInscriptions=new Vector();
    Vector errorShadows=new Vector();

    // Insert the inscriptions.
    CollectionEnumeration enum=transition.elements();
    while (enum.hasMoreElements()) {
      Object elem=enum.nextElement();
      if (elem instanceof ShadowInscription) {
	ShadowInscription inscription=(ShadowInscription)elem;
	
	CollectionEnumeration inscriptions=null;
	try {
	  inscriptions=makeInscriptions(inscription.inscr,
	    lookup,true,declaration);
	} catch (SyntaxException e) {
	  throw e.addObject(inscription);
	}

	while (inscriptions.hasMoreElements()) {
	  parsedInscriptions.addElement(inscriptions.nextElement());
	  errorShadows.addElement(inscription);
	}
      } else if (elem instanceof ShadowArc) {
	compile((ShadowArc)elem,declaration);
      } else {
	throw new SyntaxException("Unsupported transition inscription",
	  null).addObject(transition).addObject(elem);
      }
    }

    // Check for multiple uplinks.
    int uplinkCount=0;
    for (int i=0; i<parsedInscriptions.size(); i++) {
      if (parsedInscriptions.elementAt(i) instanceof UplinkInscription) {
	uplinkCount++;
      }
    }

    if (uplinkCount>1) {
       SyntaxException e=new SyntaxException(
         "Transition has more than one uplink.",null);
      for (int i=0; i<parsedInscriptions.size(); i++) {
	if (parsedInscriptions.elementAt(i) instanceof UplinkInscription) {
	  e.addObject(errorShadows.elementAt(i));
	}
      }
      throw e;     
    }

    // Insert inscriptions.
    for (int i=0; i<parsedInscriptions.size(); i++) {
      createdTransition.add(
        (TransitionInscription)parsedInscriptions.elementAt(i));
    }
  }

  Expression makeCastedOutputExpression(Class type, TypedExpression expr)
    throws SyntaxException
  {
    if (type==Types.UNTYPED) {
      return expr.getExpression();
    } else if (expr.getType()==Types.UNTYPED) {
      throw new SyntaxException(
        "Output arc expression for typed place must be typed.",
	  null);
    } else if (type==expr.getType()) {
      return expr.getExpression();
    } else if (Types.allowsLosslessWidening(expr.getType(),type)) {
      if (type.isPrimitive()) {
	return new CallExpression(expr.getExpression(),
	  new CastFunction(type));
      } else {
	return expr.getExpression();
      }
    } else {
      throw new SyntaxException(
	"Cannot losslessly convert "+
	  JavaHelper.makeTypeErrorString(expr.getType())+
	  " to "+JavaHelper.makeTypeErrorString(type)+".",
          null);
    }
  }

  Expression makeCastedInputExpression(Class type, TypedExpression expr)
    throws SyntaxException
  {
    if (type==Types.UNTYPED) {
      // We must make sure not to feed illegal
      // values into the equation.
      return JavaHelper.makeGuardedExpression(expr);
    } else if (expr.getType()==Types.UNTYPED) {
      // The expression will take all values, if we feed them
      // into it. If the expression results in illegal values,
      // there will simply be no appropriate tokens in the place.
      return expr.getExpression();
    } else if (type==expr.getType()) {
      return expr.getExpression();
    } else if (Types.allowsLosslessWidening(type,expr.getType())) {
      if (type.isPrimitive()) {
	// Although a lossless conversion to the place type
	// is not always possible, we use the same construction
	// as in lossless casts. The cast will automatically fail
	// if it is not lossless, because forward and backward
	// casts together will result in a contradiction.
	return new InvertibleExpression(expr.getExpression(),
	  new CastFunction(type),
	  new CastFunction(expr.getType()));
      } else {
	return expr.getExpression();
      }
    } else if (Types.allowsLosslessWidening(expr.getType(),type)) {
      if (type.isPrimitive()) {
	// We use the same construction as in lossless casts.
	return new InvertibleExpression(expr.getExpression(),
	  new CastFunction(type),
	  new CastFunction(expr.getType()));
      } else {
	return JavaHelper.makeGuardedExpression(expr);
      }
    } else {
      throw new SyntaxException(
	"Cannot losslessly convert "+
	  JavaHelper.makeTypeErrorString(expr.getType())+
	  " to "+JavaHelper.makeTypeErrorString(type)+
	  " or vice versa.",
          null);
    }
  }

  void compile(ShadowArc shadowArc, ParsedDeclarationNode declaration)
    throws SyntaxException
  {
    Transition t=lookup.get(shadowArc.transition);
    Place p=lookup.get(shadowArc.place);
    
    TypedExpression expr=null;
    try {
      expr=parseArcInscription(shadowArc.inscription,declaration);
      if (expr.getType()==Void.TYPE) {
        throw new SyntaxException(
          "Cannot use void expressions as arc inscriptions.",
	  null).addObject(shadowArc);
      }

      Class type=getType(shadowArc.place);
      
      Expression castedExpression;
      
      try {
	if (shadowArc.arcType==Arc.out) {
	  castedExpression=makeCastedOutputExpression(getType(shadowArc.place),
	    expr);
	} else {
	  castedExpression=makeCastedInputExpression(getType(shadowArc.place),
	    expr);
	}
      } catch (SyntaxException e) {
	throw e.addObject(shadowArc);
      }
    
      Arc arc=new Arc(p,shadowArc.arcType,castedExpression);
      arc.setTrace(shadowArc.getTrace());
      t.add(arc);
      lookup.set(shadowArc,arc);
    }
    catch (SyntaxException e) {
      throw e.addObject(shadowArc);
    }
  }

  ParsedDeclarationNode parseDeclarationNode(String inscr)
    throws SyntaxException
  {
    if (inscr!=null) {
      InscriptionParser parser=makeParser(inscr);
      try {
	return parser.DeclarationNode();
      }
      catch (ParseException e) {
	throw makeSyntaxException(e);
      }
    } else {
      return new ParsedDeclarationNode();
    }
  }

  Object parsePlaceInscription(String inscr,
    ParsedDeclarationNode declaration) throws SyntaxException
  {
    if ((inscr!=null) && !inscr.equals("")) {
      InscriptionParser parser=makeParser(inscr);
      parser.setDeclarationNode(declaration);
      try {
	return parser.PlaceTypeInscription();
      }
      catch (ParseException e) {
	// Ok, it is not a type, so it must be an initialisation
	// inscription. This is excepted.
      }

      parser=makeParser(inscr);
      parser.setDeclarationNode(declaration);
      try {
	return parser.ArcInscription();
      }
      catch (ParseException e) {
	// I really do not know what this could be.
	// I'll give out the error message of the
	// second parse, although it might be
	// appropriate to tell the user that I also
	// tried to parse a type.
	throw makeSyntaxException(e);
      }
    }

    return Types.UNTYPED;
  }

  TypedExpression parseArcInscription(String inscr,
    ParsedDeclarationNode declaration) throws SyntaxException
  {
    TypedExpression expr;

    if (inscr!=null && !inscr.equals("")) {
      InscriptionParser parser=makeParser(inscr);
      parser.setDeclarationNode(declaration);
      try {
	expr=parser.ArcInscription();
      }
      catch (ParseException e) {
	throw makeSyntaxException(e);
      }
    } else {
      expr=new TypedExpression(Tuple.class,
	new NoArgExpression(new ConstantFunction(new Tuple(0))));
    }

    return expr;
  }

  private CollectionEnumeration makeInscriptions(String str,
    ShadowLookup lookup, boolean create,
    ParsedDeclarationNode declaration) throws SyntaxException
  {
    // The input string must be parsed, it denotes either a
    // creation inscription, an uplink, a downlink, ...
    //
    // Syntax:
    //   x:new NetName          creation
    //   x:channel(...)         downlink
    //   :channel(...)          uplink
    //   x=new a.java.Type(...) creation
    //   x=y.method(...)        call
    //   guard ...              guard
    //   action x= ...          action

    InscriptionParser parser=makeParser(str);
    parser.setLookup(lookup);
    parser.setDeclarationNode(declaration);

    try {
      return parser.TransitionInscription(create);
    }
    catch (ParseException e) {
      throw makeSyntaxException(e);
    }
  }  

  ParsedDeclarationNode makeDeclarationNode(ShadowNet shadowNet)
    throws SyntaxException
  {
    // Search for declarations.
    CollectionEnumeration enum=shadowNet.elements();
    ShadowDeclarationNode node=null;
    while (enum.hasMoreElements()) {
      Object elem=enum.nextElement();
      if (elem instanceof ShadowDeclarationNode) {
        // Have we already found a declaration node?
        if (node!=null) {
	  throw new SyntaxException("Only one declaration node is allowed.",
	    null).addObject((ShadowDeclarationNode)elem);
	}
	// Remember the declaration node we found.
	node=(ShadowDeclarationNode)elem;
      }
    }

    // Compile the declaration node.
    ParsedDeclarationNode declaration=new ParsedDeclarationNode();
    if (node!=null) {
      declaration=compile((ShadowDeclarationNode)node);
    }
    return declaration;
  }

  void compile(ShadowNet shadowNet) throws SyntaxException {
    // Compile the individual net elements.
    CollectionEnumeration enum;

    // First the declaration node, if any.
    ParsedDeclarationNode declaration=makeDeclarationNode(shadowNet);

    // Second the places.
    enum=shadowNet.elements();
    while (enum.hasMoreElements()) {
      Object elem=enum.nextElement();
      if (elem instanceof ShadowPlace) {
	compile((ShadowPlace)elem,declaration);
      }
    }

    // Third the transitions.
    enum=shadowNet.elements();
    while (enum.hasMoreElements()) {
      Object elem=enum.nextElement();
      if (elem instanceof ShadowTransition) {
	compile((ShadowTransition)elem,declaration,
          lookup.getNet(shadowNet.getName()));
      }
    }
  }

  void createNets() throws SyntaxException {
    // Create new nets for every shadow net.
    // Check the uniqueness of the names.
    CollectionEnumeration enum=netSystem.elements();
    while (enum.hasMoreElements()) {
      ShadowNet shadowNet=(ShadowNet)enum.nextElement();
      if (lookup.getNet(shadowNet.getName())!=null) {
	throw new SyntaxException("Detected two nets with the same name: "+
	  shadowNet.getName()+".",null);
      }
      lookup.setNet(shadowNet.getName(),new Net(shadowNet.getName()));
    }
  }

  void createNetElements() throws SyntaxException {
    CollectionEnumeration enum=netSystem.elements();
    while (enum.hasMoreElements()) {
      compile((ShadowNet)enum.nextElement());
    }
  }

  public synchronized ShadowLookup compile(ShadowNetSystem netSystem)
    throws SyntaxException
  {
    lookup=new ShadowLookup();
    placeTypes=new Hashtable();
    this.netSystem=netSystem;
    
    createNets();
    createNetElements();

    return lookup;    
  }
}
