package de.renew.java;

import collections.*;
import de.renew.simulator.LocalVariable;

public class ParsedDeclarationNode {
  String name;
  UpdatableSeq imports;
  UpdatableMap wellKnownClasses;
  UpdatableMap variables;
  private boolean anythingDeclared;

  public ParsedDeclarationNode() {
    imports=new CircularList();
    imports.insertFirst("java.lang");
    wellKnownClasses=new HashedMap();
    variables=new HashedMap();
    name=null;
    anythingDeclared=false;
  }

  void addImport(ParsedImportDecl decl) {
    if (decl.star) {
      // Remember this qualification for lookup.
      // Which imports should have priority: early or late imports?
      imports.insertFirst(decl.name);
    } else {
      // Calculate the short name of the class.
      int pos=decl.name.lastIndexOf(".");
      String name=decl.name.substring(pos+1);
      // Map the short name to the fully qualified name.
      wellKnownClasses.putAt(name,decl.name);
    }
    anythingDeclared=true;
  }

  void addVariable(ParsedVariableDecl decl) {
    variables.putAt(decl.name,decl.clazz);
    anythingDeclared=true;
  }

  private static boolean classExists(String name) {
    try {
      Class.forName(name);
      return true;
    }
    catch (Exception e) {
      return false;
    }
  }

  private static Class classForName(String name) {
    try {
      return Class.forName(name);
    }
    catch (ClassNotFoundException e) {
      throw new RuntimeException("Could not load class.");
    }
  }

  // Return Class if class is denoted.
  // Return LocalVariable if variable is denoted.
  // Return null is no connection can be made.
  public Object interpreteName(String name) {
    // Is it a qualified name?
    if (name.indexOf(".")>=0) {
      // This could be a qualified class.
      if (classExists(name)) {
        return classForName(name);
      }
      // Don't know. Maybe it is a package specification.
      return null;      
    }

    // Do we have any declarations?
    if (!anythingDeclared) {
      // A class without package?
      if (classExists(name)) {
        return classForName(name);
      }
      // No. A language class?
      if (classExists("java.lang."+name)) {
        return classForName("java.lang."+name);
      }
      // It is a variable, even though it is not declared.
      return new LocalVariable(name);
    } else {
      // Lookup the declaration. A variable?
      if (variables.includesKey(name)) {
        return new LocalVariable(name);
      }
      // An imported class?
      if (wellKnownClasses.includesKey(name)) {
        if (classExists((String)wellKnownClasses.at(name))) {
          return classForName(name);
        }
      }
      // A class without package?
      if (classExists(name)) {
        return classForName(name);
      }
      // A class in an imported package?
      CollectionEnumeration packages=imports.elements();
      while (packages.hasMoreElements()) {
        String pckg=(String)packages.nextElement();
        if (classExists(pckg+"."+name)) {
          return classForName(pckg+"."+name);
        }
      }
      // Don't know. Maybe it is a package specification.
      return null;
    }
  }

  public Class findType(LocalVariable var) {
    if (variables.includesKey(var.name)) {
      return (Class)variables.at(var.name);
    } else {
      return de.renew.util.Types.UNTYPED;
    }
  }
}
