package de.renew.simulator;

import collections.*;
import de.renew.unify.*;
import de.renew.util.*;

class ArcOccurrence implements ExecutableOccurrence {
  TransitionInstance transitionInstance;
  PlaceInstance placeInstance;
  VariableMapper mapper;
  Variable variable;
  Arc arc;

  ArcOccurrence(Arc arc, VariableMapper mapper,
    TransitionInstance transitionInstance) 
  {
    this.arc=arc;
    this.mapper=mapper;
    this.transitionInstance=transitionInstance;
    placeInstance=transitionInstance.netInstance.getInstance(arc.place);
  }

  public CollectionEnumeration makeBinders(Searcher searcher)
    throws Impossible 
  {
    // Make sure that the place instance notifies
    // the transition instance if its marking changes,
    // because in that case the possible binding would have
    // to be rechecked.
    placeInstance.listeners.includeListener(transitionInstance);

    // For the moment, arc expressions are always early.
    // We might make them late for output arcs, but this
    // might cause significant confusion. By introducing a
    // new variable and an action statement a late evaluation
    // can always be enforced.
    variable=new Variable(
      arc.expression.startEvaluation(mapper,searcher.currentStateRecorder,
        searcher.calcChecker),
      searcher.currentStateRecorder);

    // Prepare a sequence of binders for this arc.
    UpdatableSeq binders=new LinkedList();

    if (arc.arcType==Arc.out) {
      // We must make sure that the variable associated with
      // an output arc is going to be computed before we declare
      // a transition enabled.
      searcher.calcChecker.addLateVariable(variable,
        searcher.currentStateRecorder);
    } else {
      TokenLister lister;
      // Input arcs and double arcs simply check the
      // existance of tokens during binding, so we can use
      // identical binders for them. Test arcs
      // require a different treatment, though.
      if (arc.expression.isInvertible()) {
	if (arc.arcType==Arc.test) {
	  lister=new TestTokenLister(placeInstance);
	} else {
	  lister=new InputTokenLister(placeInstance);
	}
      } else {
	lister=null;
      }

      TokenCollector collector;
      if (arc.arcType==Arc.test) {
	collector=new TestTokenCollector(placeInstance);
      } else {
	collector=new InputTokenCollector(placeInstance);
      }
      
      binders.insertFirst(new InputPlaceBinder(variable,lister,collector));
    }

    return binders.elements();
  }

  public CollectionEnumeration makeExecutables(Copier copier,
    VariableMapper copiedMapper)
  {
    // There will be no undo past this point, so we need
    // not supply a state recorder. However, we must always
    // make a copy, because the current bindings will be
    // completely rolled back. Even when we make a getValue()
    // call to the old variable, that variable might contain
    // a tuple that is taken apart later on.
    Variable copiedVariable=(Variable)copier.copy(variable);

    // Depending on the type of the arc we must create different
    // executables.
    switch (arc.arcType) {
    case Arc.in:
      return new SingletonEnumeration(
        new InputArcExecutable(placeInstance,
          copiedVariable.getValue(),arc.getTrace()));
    case Arc.test:
      return new SingletonEnumeration(
        new TestArcExecutable(placeInstance,
          copiedVariable.getValue(),arc.getTrace()));
    case Arc.out:
      return new SingletonEnumeration(
        new OutputArcExecutable(placeInstance,
          copiedVariable,arc.getTrace()));
    case Arc.both:
      return new PairEnumeration(
	new InputArcExecutable(placeInstance,
          copiedVariable.getValue(),arc.getTrace()),
        new OutputArcExecutable(placeInstance,
          copiedVariable,arc.getTrace()));
    default:
      throw new RuntimeException("Bad arc type.");
    }
  }
}
