package de.renew.simulator;

import de.renew.util.Semaphor;
import java.lang.reflect.InvocationTargetException;

// Global synchronsation order:
//   0th locally synchronized
//   1st lock SearchQueue
//   2nd lock threadLock

public class ConcurrentSimulator extends Thread
  implements Simulator, SearchQueueListener
{
  public static void main(String[] args) {
    if (args.length==0) {
      System.out.println(
        "Usage: java de.renew.simulator.ConcurrentSimulator [CLASS]");
      System.exit(0);
    }

    // Create a net.
    Net net=null;
    try {
      net=(Net)Class.forName(args[0]).
        getMethod("instance",new Class[0]).invoke(null,new Object[0]);
    }
    catch (Exception e) {
      System.out.println(args[0]+
        " isn't a valid net class as it seems.");
      if (e instanceof InvocationTargetException) {
	((InvocationTargetException)e).getTargetException().printStackTrace();
      }
      e.printStackTrace();
      System.exit(1);
    }
    
    Simulator simulator=new ConcurrentSimulator(net);
    if (simulator.getPrimaryInstance()!=null) {
      simulator.startRun();
    } else {
      System.out.println("The net did not accept the init request.");
    }
  }

  private final Searcher searcher=new Searcher();
  private final NetInstance primaryInstance;

  public ConcurrentSimulator(Net net) {
    primaryInstance=SimulatorHelper.fireInitially(net,searcher,
      new ConsoleTracer("Initially","(0) "),true);

    if (primaryInstance!=null) {
      // Run the simulation thread (me) in the background.
      setPriority(MIN_PRIORITY);
      // Start the thread.
      start();
    }
  }

  public NetInstance getPrimaryInstance() {
    return primaryInstance;
  }

  // In the concurrent simulator, you never know if
  // there might be more enabled transitions. So I am active
  // whenever the creation of the primary instance succeed.
  public boolean isActive() {
    return primaryInstance!=null;
  }

  // The following variables control the background thread.
  private long cycle=1L;
  private int desiredMode=0;
  private boolean running=true;

  private boolean stepCompleted=false;

  private boolean searchQueueIsEmpty=false;
  private boolean alreadyRegistered=false;

  private Object threadLock=new Object();
  
  public void searchQueueNonempty() {
    synchronized(threadLock) {
      searchQueueIsEmpty=false;
      alreadyRegistered=false;
      // Notify threads that may be waiting for a nonempty search queue.
      threadLock.notifyAll();
    }
  }
  
  private void registerAtSearchQueue() {
    synchronized(SearchQueue.class) {
      // Now synchronize locally to access fields safely.
      synchronized(threadLock) {
	searchQueueIsEmpty=SearchQueue.isEmpty();
	if (searchQueueIsEmpty && !alreadyRegistered) {
	  SearchQueue.insertListener(this);
	  alreadyRegistered=true;
	}
      }
    }
  }

  public synchronized void stopRun() {
    // Get thread out of the main loop, if it happens to be there.
    requestMode(0);
  }

  public synchronized boolean step() {
    // Require one more step.
    requestMode(1);

    return stepCompleted;
  }

  public synchronized void startRun() {
    // Tell the thread to start running.
    requestMode(2);
  }

  private synchronized void requestMode(int mode) {
    synchronized(threadLock) {
      // Set the mode.
      desiredMode=mode;
      // Set the running state.
      if (mode>0) {
	running=true;
      }
      // Get the thread out of the sleeping state, if it happens to be there.
      threadLock.notify();
      // Now we must wait for the thread to stop running, if we do not
      // want to let it run indefinitely.
      if (mode<2) {
	while (running) {
	  try {
	    threadLock.wait();
	  }
	  catch (InterruptedException e) {
	    // This is expected.
	  }
	}
      }
    }
  }

  public synchronized void terminateRun() {
    stop();
    synchronized (threadLock) {
      desiredMode=0;
      threadLock.notifyAll();
    }
  }

  public void run() {
    while (true) {
      // Wait until there is a reasonable chance that
      // a transition instance might be enabled.
      synchronized(threadLock) {
	while (searchQueueIsEmpty && desiredMode==2) {
	  try {
	    threadLock.wait();
	  }
	  catch (InterruptedException e) {
	    // This is expected.
	  }
	}
      }

      // Up to now we did not find any bindings.
      stepCompleted=false;

      // Try to get a transition instance from the search queue.
      // this might fail, because another thread might have emptied the
      // queue concurrently.
      Searchable searchable=SearchQueue.extract();
      while (searchable!=null && desiredMode>0) {
	// Search for a binding.
	ExecuteFinder finder=new ExecuteFinder();
	boolean bindingFound=searcher.searchOnce(finder,searchable);
	if (bindingFound) {
	  // The method search once has automatically reinserted
	  // the searchable object into the search queue.

	  // Execute the binding asynchronously.
	  finder.fireSynchronously(
	    new ConsoleTracer("Synchronously","("+(cycle++)+") "),
	    true);

	  // Make a note that we found a binding.
	  stepCompleted=true;

	  // Stop if in single step mode.
	  synchronized (threadLock) {
	    if (desiredMode==1) {
	      desiredMode=0;
	    }
	  }
	}

	// Try to find another binding.
	searchable=SearchQueue.extract();
      }

      if (searchable!=null) {
	// I have to put back the searchable
	// object that I extracted, because
	// I have not yet found time to analyse it.
	SearchQueue.possiblyActivated(searchable);
      }

      // If the queue is empty, let's register there. It's free.
      //
      // This is done even desiredMode is 0, because I would have to
      // lock threadLock to safely access desiredMode. But after locking
      // threadLock, it is forbidden to access SearchQueue.
      //
      // On the other hand, I must not lock SearchQueue all the time,
      // otherwise I might inhibit notifications from SearchQueue
      // that might keep me running.
      registerAtSearchQueue();
      
      // I must synchronize on the thread lock, otherwise
      // notifications might get lost.
      synchronized(threadLock) {
	// Did we fail to find any bindings in single step mode?
	if (desiredMode==1) {
	  // Request no further firings.
	    desiredMode=0;
	}
	
	// Did the user request a break?
	if (desiredMode==0) {
	  // I am no longer running.
	  running=false;
	  
	  // I must send a notification, just in case a stopper waits.
	  // If there are multiple stoppers, I'll notify them all.
	  threadLock.notifyAll();
	}
	
	// Now we must stick here until someone allows us to proceed.
	while (desiredMode==0) {
	  try {
	    threadLock.wait();
	  }
	  catch (InterruptedException e) {
	    // This is expected.
	  }
	}
      }
    }
  }
}


