package de.renew.simulator;

import collections.*;
import de.renew.unify.*;
import java.util.Enumeration;

public class ExecuteFinder implements Finder {
  private LinkedList executables;
  Tracer tracer;

  public ExecuteFinder() {
    executables=new LinkedList();
  }
  
  public boolean found(Searcher searcher) {
    Copier copier=new Copier();

    executables.clear();

    Enumeration transitionOccurrences=
      searcher.getOccurrences();
    while (transitionOccurrences.hasMoreElements()) {
      Occurrence occurrence=
	(Occurrence)transitionOccurrences.nextElement();
      executables.prependElements(occurrence.makeExecutables(copier));
    }
    executables.sort(new PhaseComparator());

    return true;
  }

  // The lock method must be called first. If it fails with
  // an exception, the transition cannot fire. If it succeeds,
  // it must be followed by a call to executeEarly() and then
  // by a call to executeLate().
  private void lock() throws Impossible {
    UpdatableSeq unlockables=new LinkedList();

    // Lock all early executables.
    try {
      CollectionEnumeration enum=executables.elements();
      while (enum.hasMoreElements()) {
        Executable executable=(Executable)enum.nextElement();
        if (executable instanceof EarlyExecutable) {
	  ((EarlyExecutable)executable).lock();
	  // Ok, I got the lock. Now I must make sure to
	  // treat this executable correctly: unlock or execute.
	  unlockables.insertFirst(executable);
        }
      }
    }
    catch (Impossible e) {
      // Something failed. Probably a missing token.
      CollectionEnumeration enum=unlockables.elements();
      while (enum.hasMoreElements()) {
	((EarlyExecutable)enum.nextElement()).unlock();
      }
      // I unlocked everything (in reverse order, if that matters).
      // Now I propagate the orginal exception to my caller.
      throw e;
    }
  }

  private void executeEarly(final Tracer tracer) {
    // All locks were correctly granted. Now we call all early
    // executables in their appropriate order. The execute method
    // will automatically release the executable's lock on exit.

    CollectionEnumeration enum=executables.elements();
    while (enum.hasMoreElements()) {
      Executable executable=(Executable)enum.nextElement();
      if (executable instanceof EarlyExecutable) {
	((EarlyExecutable)executable).execute(tracer);
      }
    }
  }

  private void executeLate(final Tracer tracer) {
    // All early executables were correctly executed and
    // the locks were released. Now we call all late
    // executables in their appropriate order. We do this
    // in a separate thread to avoid locking up the simulation.

    CollectionEnumeration lateEnum=executables.elements();
    while (lateEnum.hasMoreElements()) {
      Executable executable=(Executable)lateEnum.nextElement();
      if (executable instanceof LateExecutable) {
	if (((LateExecutable)executable).isLong()) {
	  de.renew.util.Detacher.detach();
	}
	try {
	  executable.execute(tracer);
	}
	catch (Exception e) {
	  //e.printStackTrace();
	  tracer.trace("Exception occurred during late phase of firing:");
	  tracer.trace(e.getMessage());
	  // Do not proceed. Some variables might have remained
	  // unbound, so it is unlikely that the following
	  // executables will succeed.
	  tracer.complete();
	  return;
	}
      }
    }
    // Everything is finished. Let me just inform the tracer.
    tracer.complete();
  }

  public boolean fireSynchronously(final Tracer tracer,
    final boolean asynchronous)
  {
    // Execute the binding that is already stored in the finder.
    try {
      lock();
      // Ok, no exception was thrown.
      executeEarly(tracer);
      // The remaining steps can be done in the background,
      // if we want to do so.
      if (asynchronous) {
        final ExecuteFinder thys=this;
	Thread lateThread=new Thread() {
	  public void run() {
	    thys.executeLate(tracer);
	    de.renew.util.Detacher.detach();
	  }
	};
	de.renew.util.Detacher.startAttached(lateThread);
      } else {
	executeLate(tracer);
      }
      
      return true;
    }
    catch (Impossible e) {
      // The transition could not fire. Probably
      // a token disappeared in the mean time.
      System.out.println("The calculated binding was disabled.");
      System.out.println("Searching for an alternative binding.");

      return false;
    }
  }
}
