package de.renew.simulator;

import collections.*;
import de.renew.unify.*;

class InputPlaceBinder implements Binder {
  Variable variable;
  TokenLister lister;
  TokenCollector collector;

  InputPlaceBinder(Variable variable,
    TokenLister lister, TokenCollector collector) 
  {
    this.variable=variable;
    this.lister=lister;
    this.collector=collector;
  }

  public int bindingBadness(Searcher searcher) {
    if (Unify.isBound(variable)) {
      // Do we need to make a copy of the desired value?
      // I think no, because the value won't be stored
      // and because it is a complete value already.
      if (collector.containsToken(variable.getValue())) {
	return 1;
      } else {
	return 0;
      }
    } else if (lister==null) {
      // We must not try to bind the variable.
      return BindingBadness.max;
    } else {
      // If we want to do it simple, we can just
      // return the cardinality of lister.getDistinctTokens(),
      // but of course this might be a very rough estimate.
      // On the other hand, making a better estimate 
      // is costly. But let's try it.
      CollectionEnumeration values=lister.getDistinctTokens();
      int count=0;
      while (values.hasMoreElements()) {
	if (Unify.isUnifiable(variable,values.nextElement())) {
	  count++;
	}
      }
      return BindingBadness.clip(count);  
    }
  }

  private boolean searchBound(Searcher searcher, Object rawValue) {
    // Here we must make copy of the value, because
    // the current value might be rolled back.
    // However, the new value must be created
    // without the possibility for rollback.
    // The value is completely bound anyway, so there
    // won't be a problem.
    Object value=Unify.copyBoundValue(rawValue);

    if (collector.containsToken(value)) {
      collector.collectToken(value);
      searcher.removeBinder(this);
      boolean success=searcher.search();
      searcher.addBinder(this);
      collector.freeToken(value);
      return success;
    } else {
      return false;
    }
  }

  public boolean bind(Searcher searcher) {
    if (Unify.isBound(variable)) {
      return searchBound(searcher,variable.getValue());
    } else {
      CollectionEnumeration values=new RandomEnumeration(
	lister.getDistinctTokens());
      while (values.hasMoreElements()) {
	Object value=values.nextElement();
	int checkpoint=searcher.currentStateRecorder.checkpoint();
	boolean success=false;
	try {
	  Unify.unify(variable,value,searcher.currentStateRecorder);
	  // Ok, unification succeeded.
	  success=searchBound(searcher,variable.getValue());
	}
	catch (Impossible e) {
	  // Unification failed.
	}
	searcher.currentStateRecorder.restore(checkpoint);
	if (success) {
	  return true;
	}
      }
    }
    return false;
  }
}
