package de.renew.simulator;

import collections.*;
import de.renew.unify.*;
import de.renew.util.Lock;
import de.renew.util.Null;
import de.renew.util.NullEnumeration;
import java.util.Enumeration;

// I will collect tokens for a single place instance. I also
// keep count of the number of tokens that are currently being
// tested in the search for a binding.
//
// Whenever my marking is changed I will notify my listeners.
//
// I allow nulls to be stored in a place. I do this by
// converting nulls to and from an internal representation using
// the class Null.

public class PlaceInstance extends java.util.Observable {
  private NetInstance netInstance;
  Place place;

  private TokenBag freeTokens;
  private TokenBag testedTokens;
  private UpdatableSeq initTokens;

  public ListenerCollection listeners;

  final long lockOrder;
  final Lock lock;

  PlaceInstance(NetInstance netInstance, Place place)
    throws Impossible
  {
    this.netInstance=netInstance;
    this.place=place;

    lockOrder=de.renew.util.Orderer.getTicket();
    lock=new Lock();

    freeTokens=new TokenBag();
    testedTokens=new TokenBag();

    listeners=new ListenerCollection();

    // Let's calculate the initial bag of tokens.
    initTokens=new LinkedList();
    CollectionEnumeration inscriptions=place.inscriptions.elements();
    if (inscriptions.hasMoreElements()) {
      do {
	InitialMarking initialMarking=
	  (InitialMarking)inscriptions.nextElement();
	Object value=
	  initialMarking.getInitialMarking();
	initTokens.insertFirst(Null.convert(value));
      } while(inscriptions.hasMoreElements());
    }
  }

  public String toString() {
    return netInstance.toString()+"."+place.toString();
  }

  // I should rather return a copy of
  // my token bag. This is dangerous !!! 
  public TokenBag getTokens() {
    return freeTokens;
  }

  CollectionEnumeration getDistinctTokens() {
    return Null.convertEnumeration(freeTokens.uniqueElements());
  }

  // Here we have to include the already tested tokens.
  CollectionEnumeration getDistinctTestableTokens() {
    UpdatableSet individualTokens=new HashedSet();
    individualTokens.includeElements(freeTokens.uniqueElements());
    individualTokens.includeElements(testedTokens.uniqueElements());

    return Null.convertEnumeration(individualTokens.elements());
  }

  boolean containsToken(Object token) {
    return freeTokens.includes(Null.convert(token));
  }
  // Here consider only already tested tokens.
  boolean containsTestedToken(Object token) {
    return testedTokens.includes(Null.convert(token));
  }
  // Here we consider both kinds of tokens.
  boolean containsTestableToken(Object token) {
    // nulls will be converted by the called methods.
    return containsToken(token) || containsTestedToken(token);
  }

  void removeToken(Object token) {
    // The following code is synchronized to avoid concurrent
    // update requests. This does not preclude all conflicts
    // that such a notification might cause. The observers must
    // still take some care although it might work now.
    // NOTE: CAREFUL EXAMINATION REQUIRED!!!
    synchronized (freeTokens) {
      freeTokens.removeOneOf(Null.convert(token));
      // Observers should be notified in any case, although listeners
      // (which are responsible for keeping track of enabled
      // transitions) must only be notified if a marking increases.
      setChanged();
      notifyObservers(freeTokens);
    }
  }

  void insertToken(Object token) {
    // The following code is synchronized to avoid concurrent
    // update requests. This does not preclude all conflicts
    // that such a notification might cause. The observers must
    // still take some care although it might work now.
    // NOTE: CAREFUL EXAMINATION REQUIRED!!!
    synchronized (freeTokens) {
      freeTokens.add(Null.convert(token));
      listeners.notifyChange();
      setChanged();
      notifyObservers(freeTokens);
    }
  }

  // A reserve acts almost like remove, but it does not 
  // notify the observers that are attached to this 
  // place instance, because the removal of tokens
  // is only temporary.
  void reserveToken(Object token) {
    freeTokens.removeOneOf(Null.convert(token));
  }

  // An unreserve acts almost like insert, but it
  // does not notify the listeners, because it is
  // supposed to return only tokens that were
  // already present when the listeners accessed
  // this object. Observers aren't notified, either.
  void unreserveToken(Object token) {
    freeTokens.add(Null.convert(token));
  }

  // Multiple tests on the same token do not remove
  // the token multiple times. Instead, the token is
  // removed once, and the number of removals is recorded
  // by a multiple insertion into the bag of tested tokens.
  void testToken(Object token) {
    token=Null.convert(token);
    if (!testedTokens.includes(token)) {
      freeTokens.removeOneOf(token);
    }
    testedTokens.add(token);
  }

  // Undo a testToken call. Once the last token is returned,
  // a token is reinserted into the bag of free tokens.
  void untestToken(Object token) {
    token=Null.convert(token);
    testedTokens.removeOneOf(token);
    if (!testedTokens.includes(token)) {
      freeTokens.add(token);
    }
  }

  /**
   * My net will call this method if its creation is confirmed.
   * I will free all tokens that I kept during my initialisation.
   * You must call this method at most once.
   **/
  void createConfirmation(Tracer tracer) {
    // The following code is synchronized to avoid concurrent
    // update requests. This does not preclude all conflicts
    // that such a notification might cause. The observers must
    // still take some care although it might work now.
    // NOTE: CAREFUL EXAMINATION REQUIRED!!!
    synchronized (freeTokens) {
      CollectionEnumeration enum=initTokens.elements();
      // Anything to do?
      if (enum.hasMoreElements()) {
	// Yes, move all tokens to their destination.
	do {
	  Object token=enum.nextElement();
	  tracer.trace("    Putting "+Null.convert(token)+
	    " into "+this);
	  freeTokens.add(token);
	} while (enum.hasMoreElements());
	
	// Just notify the listeners once.
	listeners.notifyChange();
	setChanged();
	notifyObservers(freeTokens);
      }
      
      // Null the field to allow garbage collection.
      initTokens=null;
    }
  }
}

