package de.renew.simulator;

import de.renew.util.EmptyEnumeration;
import collections.*;

/**
 * A random search queue returns possibly enabled searchable
 * objects in a random order. This is useful for running many
 * simulations that explore the possible behaviour of a net.
 *
 * The implementation is based on an array. During the
 * extraction, a random element is removed and the 
 * void position is filled in with the very last element.
 *
 * @author Olaf Kummer
 **/

class RandomSearchQueue implements SearchQueueData {
  private Searchable[] elements;
  private int size;
  private UpdatableSet isUnchecked;

  RandomSearchQueue() {
    init(new EmptyEnumeration());
  }

  private void setCapacity(int capacity) {
    Searchable[] newElements=new Searchable[capacity];
    System.arraycopy(elements,0,newElements,0,size);
    elements=newElements;
  }

  private void ensureCapacity(int capacity) {
    if (capacity>elements.length) {
      // Ensure that the capacity is at least doubled
      // each time.
      if (capacity<2*elements.length) {
	capacity=2*elements.length;
      }
      setCapacity(capacity);
    }
  }

  private void limitCapacity() {
    // Ensure that the capacity can be at least halved
    // each time. Also, there must remain enough elements
    // that no enlargement is required in the near future.
    // It makes no sense to resize very small arrays,
    // because the creation of a new array is very expensive.
    if (size>=8 && size*4<elements.length) {
      setCapacity(size*2);
    }
  }

  public void possiblyActivated(Searchable transition) {
    if (!isUnchecked.includes(transition)) {
      ensureCapacity(size+1);
      elements[size++]=transition;
      isUnchecked.include(transition);
    }
  }

  public boolean isEmpty() {
    return size==0;
  }

  public Searchable extract() {
    if (isEmpty()) {
      return null;
    }
    // Choose a random element.
    int i=(int)(Math.random()*size);
    // Guard against rounding errors.
    if (i>=size) { i=size-1; }
    // Remember element.
    Searchable transition=elements[i];
    // Pull in last element.
    elements[i]=elements[size-1];
    // Null the field to allow garbage collection.
    elements[size-1]=null;
    size=size-1;
    limitCapacity();

    isUnchecked.exclude(transition);

    return transition;
  }

  public CollectionEnumeration elements() {
    return isUnchecked.elements();
  }

  public void init(CollectionEnumeration enum) {
    elements=new Searchable[8];
    size=0;
    isUnchecked=new HashedSet();
    
    while (enum.hasMoreElements()) {
      Searchable searchable=(Searchable)enum.nextElement();
      possiblyActivated(searchable);
      isUnchecked.include(searchable);
    }
  }
}

