package de.renew.simulator;

import collections.*;
import java.util.Vector;
import java.util.Enumeration;
import de.renew.util.Semaphor;
import de.renew.unify.*;

public class Searcher {
  // A searcher keeps track of various possibilities to
  // refine a partially known binding. By default the
  // possibility with the least number of branches to examine
  // is chosen, but this construction allows a simple exchange
  // of algorithms.
  //
  // The algorithm may proceed to bind tokens to arc expressions,
  // downlinks to individual transitions, or
  // transitions to arc expressions.

  private boolean bindingFound;
  private Finder finder;

  private UpdatableSet binders;
    // UpdatableSet<Binder>
  private UpdatableSet occurrences;
    // UpdatableSet<Occurrence>

  public final CalculationChecker calcChecker;

  public final StateRecorder currentStateRecorder;

  // A lock that ensures that no two searchers are
  // active concurrently. This is required as long as
  // token access conflicts are not resolved explicitly.
  //
  // Concurrent access is especially dangerous, when a
  // net tries to invoke the searcher directly.
  // For the moment we ignore this possibility.
  static Semaphor searchLock;
  static {
    // Create the lock.
    searchLock=new Semaphor();
    // Grant one concurrent access.
    searchLock.V();
  }

  public Searcher() {
    currentStateRecorder=new StateRecorder();

    binders=new HashedSet();
    occurrences=new HashedSet();

    calcChecker=new CalculationChecker();
  }

  // Maintain a collection of transition occurrences.
  public void addOccurrence(Occurrence occurrence) {
    occurrences.include(occurrence);
  }

  public void removeOccurrence(Occurrence occurrence) {
    occurrences.exclude(occurrence);
  }

  public CollectionEnumeration getOccurrences() {
    return occurrences.elements();
  }

  // Find the binder with the minimum binding badness
  // Return null, if no binder wants to try.
  private Binder selectBestBinder() {
    Binder bestBinder=null;
    int bestBadness=BindingBadness.max;
    Enumeration enum=binders.elements();
    while (enum.hasMoreElements()) {
      Binder binder=(Binder)enum.nextElement();
      int badness=binder.bindingBadness(this);
      if (badness<bestBadness) {
	bestBinder=binder;
	bestBadness=badness;
      }
    }
    return bestBinder;
  }

  public boolean search() {
    boolean result;
    if (!binders.isEmpty()) {
      Binder binder=selectBestBinder();
      if (binder!=null) {
	int checkpoint=currentStateRecorder.checkpoint();
        result=binder.bind(this);
	currentStateRecorder.restore(checkpoint);
      } else {
	// The remaining binders cannot proceed.
	// This is probably due to an unbound variable.
	result=false;
      }
    } else {
      try {
	calcChecker.checkConsistency();
	bindingFound=true;
	result=finder.found(this);
      }
      catch (Impossible e) {
	result=false;
      }
    }
    return result;
  }

  public void addBinder(Binder binder) {
    binders.include(binder);
  }
  public void removeBinder(Binder binder) {
    binders.exclude(binder);
  }
  public void addBinders(Enumeration newBinders) {
    binders.includeElements(newBinders);
  }
  public void removeBinders(Enumeration oldBinders) {
    binders.removeElements(oldBinders);
  }

  public void startSearch() {
    // Make sure that the current state recorder does not waste
    // space with now obsolete bindings.
    currentStateRecorder.restore();

    // Ensure that the temoprary containers are initially empty.
    // If they are not, something went wrong previously,
    // so we report en error.
    if (!occurrences.isEmpty() || !binders.isEmpty()) {
      throw new RuntimeException("Searcher was not in idle state "+
	"at the start of a search.");
    }
  }

  public boolean searchLoop(Finder finder) {
    // Perpare a new search that starts from scratch.
    startSearch();

    // Remember the finder.
    this.finder=finder;

    // Make a list of all the transition instances
    // that could fire, even if they were rejected by the
    // finder.
    UpdatableSeq wasActivated=new LinkedList();

    // Nothing found so far.
    boolean result=false;

    // Try all transitions that might be enabled.
    // If a successful binding was found (though not
    // neccessarily accepted), requeue this transition
    // at the end of the search.
    // If no bindings are possible, forget about this 
    // transition. A change to a relevant place will
    // trigger its reinsertion.
    while ((!result) && (!SearchQueue.isEmpty())) {
      // Take a note that no binding has been found so far.
      bindingFound=false;

      Searchable searchable=SearchQueue.extract();

      // Is this searchable object activated?
      result=searchable.search(this);

      // We must check for a reinsertion of this transition
      // into the list of activated elements, regardless whether this
      // transition instance is actually selected by the binder.
      if (bindingFound) {
	// Ok, reinsert this transition.
	wasActivated.insertFirst(searchable);
      }
    }

    // Reinsert the activated transitions into the search list.
    // The intermediate list ensured the termination of the inner loop,
    // when a binder does not accept any of the proposed bindings.
    CollectionEnumeration enum=wasActivated.elements();
    while (enum.hasMoreElements()) {
      SearchQueue.possiblyActivated(
        (Searchable)enum.nextElement());
    }

    // Forget the finder.
    this.finder=null;

    return result;
  }

  // Try to find a binding for a searchable. If neccessary,
  // reinsert the searchable into the search queue.
  public boolean searchOnce(Finder finder, Searchable searchable) {
    // Perpare a new search that starts from scratch.
    startSearch();

    // Remember the finder.
    this.finder=finder;

    // Take a note that no binding has been found so far.
    bindingFound=false;
    
    // Is this searchable activated?
    boolean result=searchable.search(this);

    // We must check for a reinsertion of this searchable
    // into the list of activated elements, regardless whether this
    // searchable instance is actually selected by the binder.
    if (bindingFound) {
      // Ok, reinsert this searchable.
      SearchQueue.possiblyActivated(searchable);
    }

    // Forget the finder.
    this.finder=null;

    return result;
  }

  public boolean initiatedSearch(NetInstance netInstance,
    String name, Variable params, boolean isOptional, Finder finder)
  {
    // Perpare a new search that starts from scratch.
    startSearch();

    // Remember the finder.
    this.finder=finder;

    // We want to find a transition within the specified
    // net instance that can fire. This method is
    // usually called to generate the very first tokens
    // in a net. In a sense, this resembles the main(args)
    // call of the runtime environment.
    //
    // If the argument isOptional is true, the invoked net
    // need not provide an appropriate channel at all. But if it does,
    // the synchronisation must succeed.
    Variable netVariable=new Variable();
    try {
      Unify.unify(netVariable,netInstance,currentStateRecorder);
    }
    catch (Impossible e) {
      throw new RuntimeException("Unification failed unexpectedly.");
    }
    Binder initialBinder=
      new NetChannelBinder(netVariable,name,params,isOptional,null);
    addBinder(initialBinder);
    boolean result=search();
    removeBinder(initialBinder);

    finder=null;
    currentStateRecorder.restore();

    return result;
  }
}

