package de.renew.simulator;

import java.lang.reflect.InvocationTargetException;
import de.renew.unify.*;

public class SequentialSimulator
  implements Simulator, Runnable
{
  public static void main(String[] args) {
    if (args.length==0) {
      System.out.println(
        "Usage: java de.renew.simulator.SequentialSimulator [CLASS]");
      System.exit(0);
    }

    // Create a net.
    Net net=null;
    try {
      net=(Net)Class.forName(args[0]).
        getMethod("instance",new Class[0]).invoke(null,new Object[0]);
    }
    catch (Exception e) {
      System.out.println(args[0]+
        " isn't a valid net class as it seems.");
      if (e instanceof InvocationTargetException) {
	((InvocationTargetException)e).getTargetException().printStackTrace();
      }
      e.printStackTrace();
      System.exit(1);
    }
    
    
    Simulator simulator=new SequentialSimulator(net);
    if (simulator.getPrimaryInstance()!=null) {
      simulator.startRun();
    } else {
      System.out.println("The net did not accept the init request.");
    }
  }

  private ExecuteFinder finder=new ExecuteFinder();
  private Searcher searcher=new Searcher();
  private NetInstance primaryInstance;
  private boolean bindingFound;
  private boolean stepFired=false;;

  private boolean wantBreak=false;
  private Thread runThread=null;

  public SequentialSimulator(Net net) {
    primaryInstance=SimulatorHelper.fireInitially(net,searcher,
      new ConsoleTracer("Initially"),false);
    
    if (primaryInstance!=null) {
      // If possible, prepare a new binding.
      findNewBinding();
    }
  }

  public NetInstance getPrimaryInstance() {
    return primaryInstance;
  }

  public boolean isActive() {
    return bindingFound;
  }

  public synchronized void startRun() {
    // Only create a new thread if none is currently running.
    if (runThread==null) {
      // Create a new thread that will run the simulation.
      runThread=new Thread(this);
      // Run the thread in the background.
      runThread.setPriority(Thread.MIN_PRIORITY);
      // Start the thread.
      runThread.start();
    }
  }

  // Gently stop the simulation.
  // This may take a while, because the current search for
  // a activated binding must be completed first.
  // After the completion of this method, no thread is running
  // and runThread==null.
  public synchronized void stopRun() {
    while (runThread!=null) {
      wantBreak=true;
      try {
	wait();
      }
      catch (InterruptedException e) {
	// This is expected. But I cannot be sure that
	// I woke up from a notification.
      }
    }
  }

  // Stop the simulation the hard way.
  // This may corrupt the simulation state.
  public synchronized void terminateRun() {
    if (runThread!=null) {
      runThread.stop();
      runThread=null;
      wantBreak=false;
      notifyAll();
    }
  }

  private void fire(Tracer tracer) {
    stepFired=bindingFound;
    if (bindingFound) {
      finder.fireSynchronously(tracer,false);
      findNewBinding();
    }
  }

  // To exclude concurrent access to the searcher, this method may only
  // be called from the running thread or from a
  // synchronized method that has ensured that no thread is running.
  // This method is not synchronized, because it could
  // lock up termination requests.
  // If there are no more enabled transitions, this method simply
  // returns without effect.
  private void findNewBinding() {
    bindingFound=searcher.searchLoop(finder);
    if (!bindingFound) {
      System.out.println("No more enabled transitions.");
    }
  }

  // This method must be synchronized to avoid concurrent
  // accesses to restart another thread after the stop.
  public synchronized boolean step() {
    if (runThread!=null) {
      // Just stop the current thread. It will make another step
      // anyway.
      stopRun();
    } else {
      // Try to make another step.
      fire(new ConsoleTracer("Synchronously"));
    }
    return stepFired;
  }

  // This method is called in threads that are generated on each
  // run request. The simulator class cannot inherit from thread,
  // because this method might be run multiple times.
  public void run() {
    // Make steps until deadlock or until a break is requested.
    while (bindingFound && !wantBreak) {
      // Grant other threads a chance to run.
      // Only needed on systems without preemptive multitasking.
      Thread.yield();
      // Fire and search for a new binding.
      fire(new ConsoleTracer("Synchronously"));
    }
    // Make sure that runThread and wantBreak are updated
    // atomically.
    synchronized(this) {
      // Signal the end of the run.
      runThread=null;
      // Clear the break request.
      wantBreak=false;
      // Notify waiting threads of break.
      notifyAll();
    }
  }
}
