package de.renew.simulator;

import collections.*;
import de.renew.unify.Impossible;

/**
 * I represent an instance of a transition. Unlike a naked transition
 * I am tied to a net instance. I provide my name to the outside world
 * and keep track of those objects who like to be notified if I become
 * enabled.
 *
 * Other objects will register me as a listener of objects that might
 * enable me. I would like to be registered as a listener with
 * all objects that might enable me, but this is beyond my control.
 * So take a little care when you use me.
 *
 * @author Olaf Kummer
 **/
class TransitionInstance implements Listener, Searchable {
  /**
   * My net instance to which I belong.
   **/
  NetInstance netInstance;
  /**
   * The transition that serves as a patern for me.
   **/
  Transition transition;

  /**
   * A collection of listeners that I must notify when I become
   * enabled.
   **/
  ListenerCollection listeners;

  /**
   * I (a transition instance) am created. I remember
   * the neccessary information and prepare myself for
   * possible listeners.
   *
   * @param netInstance
   *   the net instance that needs to have an instance of some transition
   * @paramtransition
   *   the transition that must be instantiated
   **/
  TransitionInstance(NetInstance netInstance,
    Transition transition)
  {
    this.netInstance=netInstance;
    this.transition=transition;

    // Note that a notification is only provided if
    // a transition becomes potentially activated.
    // Is that ok?
    listeners=new ListenerCollection();
  }

  /**
   * I will use the name of my net instance plus the name of
   * my transition to build my own name.
   *
   * @return my name
   **/
  public String toString() {
    return netInstance.toString()+"."+transition.toString();
  }

  /**
   * Somebody will call this method to tell me that I might
   * be enabled. I will tell the search queue about it, if I am
   * spontaneous, and I will notify my listeners.
   */
  public void notifyChange() {
    // Notify the searcher, if this transition is
    // spontaneous.
    if (transition.isSpontaneous()) {
      SearchQueue.possiblyActivated(this);
    }
    // Notify the listeners, if any.
    listeners.notifyChange();
  }

  /**
   * My net will call this method if it creation is confirmed.
   * Just to be safe, I will register myself at the search queue
   * if I am spontaneous.
   **/
  void createConfirmation() {
    if (transition.isSpontaneous()) {
      SearchQueue.possiblyActivated(this);
    }
  }

  /**
   * This method is called if somebody wants me to start a
   * search for an activated binding. I initiate the search and
   * let the searcher do the rest.
   *
   * @param searcher
   *   the searcher that is supposed to actually perform the search.
   **/
  public boolean search(Searcher searcher) {
    boolean result;
    try {
      // Prepare a new instance to check.
      TransitionOccurrence occurrence=
	new TransitionOccurrence(this,null,searcher);
      
      // No exception was thrown, so this transition
      // may be checked.
      searcher.addOccurrence(occurrence);
      searcher.addBinders(occurrence.getBinders());
      result=searcher.search();
      searcher.removeBinders(occurrence.getBinders());
      searcher.removeOccurrence(occurrence);
    }
    catch (Impossible e) {
      // I had to refuse the search request completely.
      // Probably there is a cyclic dependency of some
      // calculations. This should really be detected during
      // the syntax check.
      result=false;
    }

    searcher.currentStateRecorder.restore(); // no bindings sensible

    return result;  
  }
}
