package de.renew.simulator;

import collections.*;
import de.renew.unify.*;

/**
 * Upon evaluation I will spawn the evaluation of several
 * supexpressions. I will then compose a tuple from the resulting values.
 *
 * @author Olaf Kummer
 **/
public class TupleExpression implements Expression {
  /**
   * A tuple with the expressions that I must evaluate.
   * Hmm, honestly, is there a reason why this must be
   * stored in a tuple? Inefficient. Probably that's just a
   * symmetry consideration.
   **/
  Tuple expressions;

  /**
   * I (a tuple expression) am created.
   * 
   * @param expressions
   *   a tuple that holds the expressions that must be evaluated
   **/
  public TupleExpression(Tuple expressions) {
    this.expressions=expressions;
  }

  /**
   * Create a new tuple expression from exactly one
   * expression. Unary tuples are not used very often,
   * but they do occur sometimes.
   *
   * @param expr1
   *   the first expression
   **/
  public TupleExpression(Expression expr1) {
    expressions=new Tuple(new Object[]{expr1},null);
  }

  /**
   * Create a new tuple expression from exactly two
   * expressions. Because pairs are so common, it seems
   * desirable to have a short notation for them.
   *
   * @param expr1
   *   the first expression
   * @param expr2
   *   the second expression
   **/
  public TupleExpression(Expression expr1, Expression expr2) {
    expressions=new Tuple(new Object[]{expr1,expr2},null);
  }

  /**
   * Create a new tuple expression from exactly two
   * expressions. Because triples are still farily common, it seems
   * desirable to have a short notation for them.
   *
   * @param expr1
   *   the first expression
   * @param expr2
   *   the second expression
   * @param expr3
   *   the third expression
   **/
  public TupleExpression(Expression expr1, Expression expr2,
    Expression expr3)
  {
    expressions=new Tuple(new Object[]{expr1,expr2,expr3},null);
  }

  /**
   * Am I invertible? Of course I am.
   *
   * @return true
   **/
  public boolean isInvertible() {
    return true;
  }

  /**
   * There is really not much difference between evaluation
   * and registration. Only, I have to evaluate the subexpressions
   * accordingly.
   *
   * @param mapper
   *   a variable mapper which I must pass to the subexpressions
   * @param recorder
   *   a state recorder which I must pass to the subexpressions
   * @param checker
   *   a calculation checker which I must pass to the subexpressions
   * @param forEvaluation
   *   true, if we are trying an evaluation
   * @return the constructed tuple, possibly incomplete
   **/
  private Tuple makeResultTuple(VariableMapper mapper,
    StateRecorder recorder, CalculationChecker checker,
    boolean forEvaluation) throws Impossible
  {
    int n=expressions.getArity();
    Tuple tuple=new Tuple(n);

    for (int i=0;i<n;i++) {
      Expression expr=(Expression)expressions.getComponent(i);
      Object obj;
      if (forEvaluation) {
	obj=expr.startEvaluation(mapper,recorder,checker);
      } else {
	obj=expr.registerCalculation(mapper,recorder,checker);
      }
      Unify.unify(tuple.getComponent(i),obj,recorder);
    }

    return tuple;
  }

  /**
   * I will evaluate the subexpressions and put the results
   * into a tuple.
   *
   * @param mapper
   *   a variable mapper which I must pass to the subexpressions
   * @param recorder
   *   a state recorder which I must pass to the subexpressions
   * @param checker
   *   a calculation checker which I must pass to the subexpressions
   * @return the constructed tuple, possibly incomplete
   **/
  public Object startEvaluation(VariableMapper mapper, StateRecorder recorder, 
    CalculationChecker checker) throws Impossible
  {
    return makeResultTuple(mapper,recorder,checker,true);
  }

  /**
   * I will register the subexpressions and put the results
   * into a tuple.
   *
   * @param mapper
   *   a variable mapper which I must pass to the subexpressions
   * @param recorder
   *   a state recorder which I must pass to the subexpressions
   * @param checker
   *   a calculation checker which I must pass to the subexpressions
   * @return the constructed tuple, possibly incomplete
   **/
  public Object registerCalculation(VariableMapper mapper,
    StateRecorder recorder, CalculationChecker checker) throws Impossible
  {
    return makeResultTuple(mapper,recorder,checker,false);
  }
}

