package de.renew.unify;

import collections.*;

abstract class Carrier {
  UpdatableSet backlinks;

  boolean recorderUsed;

  Object[] references;
  int arity; 

  boolean complete; // Are no unknowns part of this object?
  boolean bound; // Is this object bound to a real value?

  Carrier(Object[] references, StateRecorder recorder) {
    backlinks=new HashedSet();

    recorderUsed=false;

    this.references=references;
    arity=references.length;

    for (int i=0;i<arity;i++) {
      Object o=references[i];
      if (o instanceof Carrier) {
	((Carrier)o).addBacklink(this,recorder);
      }
    }

    complete=calculateComplete();
    if (complete) {
      bound=calculateBound();
    } else {
      bound=false;
    }
  }

  boolean calculateComplete() {
    for (int i=0;i<arity;i++) {
      if (!Unify.isComplete(references[i])) {
	return false;
      }
    }
    return true;
  }

  boolean calculateBound() {
    for (int i=0;i<arity;i++) {
      if (!Unify.isBound(references[i])) {
	return false;
      }
    }
    return true;
  }

  // Hide the normal equal and hashCode methods.
  // Because cariers can change their values, it would
  // be desastrous to insert them into a hashtable.
  // Tuples and variables will redeclare these methods
  // in a sensible way.
  public boolean equal(Object that) {
    throw new RuntimeException(
      "Somebody compared an incomplete unifiable object.");
  }
  public int hashCode() {
    throw new RuntimeException(
      "Somebody took the hash code of an incomplete unifiable object.");
  }

  void checkRecorder(StateRecorder recorder) {
    if (recorder==null) {
      if (recorderUsed) {
	throw new RuntimeException(
          "Permanent action after undoable action. Strange.");
      }
    } else {
      recorderUsed=true;
    }
  }

  void addBacklink(Carrier c, StateRecorder recorder) {
    final Object wrapped=new IdentityWrapper(c);
    if (!backlinks.includes(wrapped)) {
      backlinks.include(wrapped);
      checkRecorder(recorder);
      if (recorder!=null) {
	recorder.record(new StateRestorer() {
	  public void restore() {
	    backlinks.exclude(wrapped);
	  }
	});
      }
    }
  }

  void updateBacklinked(Object value, UpdatableSet listeners,
    StateRecorder recorder) throws Impossible
  {
    // I will tell all my backlinked objects that the given
    // value is now relevant instead of myself. (Maybe I *am* the
    // new object, but who knows.)
    //
    // I assume that no new backlinks will be added to myself
    // during this procedure.
    CollectionEnumeration enum=backlinks.elements();
    while (enum.hasMoreElements()) {
      IdentityWrapper wrapper=(IdentityWrapper)enum.nextElement();
      Carrier carrier=(Carrier)wrapper.getObject();
      carrier.update(this,value,listeners,recorder);
    }
  }

  void update(final Carrier old, final Object value,
    UpdatableSet listeners, StateRecorder recorder) throws Impossible
  {
    // This is used to be a plausibility check. But it
    // was overly strict. The completion of a subobject might be
    // indicated multiple times. Scenario: A=[B,C], B=[C]
    // and C becomes complete. Now C notifies B, which becomes
    // complete and notifies A in turn, which is complete now, too.
    // Now the notification from C to A (C occurs in A)
    // remains to be done, but A is already complete.

    //if (complete) {
    //  throw new RuntimeException(
    //    "A complete carrier was notified again. Strange.");
    //}

    // Only update the link if the value has changed.
    if (old!=value) {
      // We must go through all the references to find potential updates.
      // This seems inefficient, but usually the list will be short,
      // and this way we do not have to save which positions
      // need to be notified.
      for (int i=0;i<arity;i++) {
	final int iFinal=i;
	if (references[iFinal]==old) {
	  checkRecorder(recorder);
	  if (recorder!=null) {
	    recorder.record(new StateRestorer() {
	      public void restore() {
		references[iFinal]=old;
	      }
	    });
	  }
	  references[iFinal]=value;
	  if (value instanceof Carrier) {
	    ((Carrier)value).addBacklink(this,recorder);
	  }
	}
      }
    }
    // Maybe I am complete by now?
    if (calculateComplete()) {
      checkRecorder(recorder);
      if (recorder!=null) {
	recorder.record(new StateRestorer() {
	  public void restore() {
	    complete=false;
	}
	});
      }
      complete=true;
      // Maybe I am even bound?
      if (calculateBound()) {
	checkRecorder(recorder);
	if (recorder!=null) {
	  recorder.record(new StateRestorer() {
	    public void restore() {
	      bound=false;
	    }
	  });
	}
	bound=true;
      }
      // I any case, I have a notification to send to my backlinked objects.
      // I am notifying with myself as the new value.
      updateBacklinked(this,listeners,recorder);
    }
  }

  void occursCheck(Object that, UpdatableSet visited)
    throws Impossible
  {
    // Do I cause a cycle?
    if (this==that) {
      throw new Impossible();
    }
    // Did I check myself earlier?
    IdentityWrapper thisWrapper=new IdentityWrapper(this);
    if (visited.includes(thisWrapper)) {
      return;
    }
    // I do not want to check me again.
    visited.include(thisWrapper);
    // Now let's test all the referenced objects.
    for (int i=0;i<arity;i++) {
      if (references[i] instanceof Carrier) {
	((Carrier)references[i]).occursCheck(that,visited);
      }
    }
  }
}
