package de.renew.unify;

import collections.*;

public final class Unify {
  private Unify(){}

  static Object[] cleanupArray(Object[] orgArray) {
    Object[] initArray=new Object[orgArray.length];
    for (int i=0;i<orgArray.length;i++) {
      if (orgArray[i] instanceof Variable) {
	initArray[i]=((Variable)orgArray[i]).getValue();
      } else {
	initArray[i]=orgArray[i];
      }
    }
    return initArray;
  }

  static Object[] makeInitArray(Object initValue) {
    Object[] initArray=new Object[1];
    if (initValue instanceof Variable) {
      initArray[0]=((Variable)initValue).getValue();
    } else {
      initArray[0]=initValue;
    }
    return initArray;
  }

  static Object[] makeUnknownArray(int arity) {
    Object[] unknowns=new Object[arity];
    for (int i=0;i<arity;i++) {
      unknowns[i]=new Unknown();
    }
    return unknowns;
  }

  // The next two methods probe the status of a unifiable object.
  // An object is complete when no further unification
  // within its scope can change its state, i.e. it is
  // fuilly calculateable.
  // An object is bound if it is fully calculated.
  //
  // An object may be complete and not bound, if it depends on
  // caculations registered with the checker.
  // A bound object if always complete.

  static boolean isComplete(Object o) {
    if (o instanceof Carrier) {
      // Let's ask the object.
      return ((Carrier)o).complete;
    } else {
      // Ordinary objects are always complete.
      return true;
    }
  }

  public static boolean isBound(Object o) {
    if (o instanceof Carrier) {
      // Let's ask the object.
      return ((Carrier)o).bound;
    } else {
      // Ordinary objects are always bound.
      return true;
    }
  }

  static private void unifyUnknown(Unknown left, Object right,
    StateRecorder recorder) throws Impossible
  {
    // If the other object is a carrier too, we must
    // make an occurs check.
    if (right instanceof Carrier) {
      ((Carrier)right).occursCheck(left,new HashedSet());
    }

    // Prepare a set of listeners that will hold the
    // listeners to invoke.
    UpdatableSet listeners=new HashedSet();

    // Make the update.
    left.updateBacklinked(right,listeners,recorder);

    // Notify all listeners.
    CollectionEnumeration enum=listeners.elements();
    while(enum.hasMoreElements()) {
      ((Notifiable)enum.nextElement()).boundNotify(recorder);
    }
  }

  static private void unifyTuples(Tuple left, Tuple right,
    StateRecorder recorder) throws Impossible
  {
    int arity=left.getArity();
    if (right.getArity()!=arity) {
      throw new Impossible();
    }
    for (int i=0;i<arity;i++) {
      unify(left.getComponent(i),right.getComponent(i),recorder);
    }
  }

  static public void unify(Object left, Object right, StateRecorder recorder)
    throws Impossible
  {
    // Remove variables, because they only wrappers
    // for the outside world.
    if (left instanceof Variable) {
      left=((Variable)left).getValue();
    }
    if (right instanceof Variable) {
      right=((Variable)right).getValue();
    }

    // Now do the unification.
    if (left==right) {
      return;
    } else if (left instanceof Unknown) {
      unifyUnknown((Unknown)left,right,recorder);
    } else if (right instanceof Unknown) {
      unifyUnknown((Unknown)right,left,recorder);
    } else if (left instanceof Tuple) {
      if (right instanceof Tuple) {
	unifyTuples((Tuple)left,(Tuple)right,recorder);
      } else {
	throw new Impossible();
      }
    } else if (left==null) {
      throw new Impossible(); // We already checked for identity.
    } else if (right==null) {
      throw new Impossible(); // We already checked for identity.
    } else {
      if (!left.equals(right)) {
	throw new Impossible();
      }
    }    
  }

  // Provide a state recorder that can be used in short term
  // undo operations.
  private static StateRecorder tempRecorder=new StateRecorder();

  // Check whether the arguments are unfiable.
  // At the moment also send notifications
  // so that everything is checked. But maybe it is
  // more sensible to inhibit notification during this test?
  static public boolean isUnifiable(Object left, Object right) {
    boolean result=false;
    synchronized(tempRecorder) {
      try {
	unify(left,right,tempRecorder);
	result=true;
      }
      catch (Impossible e) {
	// This is expected.
      }
      tempRecorder.restore();
    }
    return result;
  }

  // Provide a copier that can be used to copy bound values.
  private static Copier tempCopier=new Copier();
  static public Object copyBoundValue(Object val) {
    if (!isBound(val)) {
      throw new RuntimeException("To copy unbound values, use a copier.");
    }
    return tempCopier.copy(val);
  }

  // Ensure that target==function(source) after unification.
  // The equation is guaranteed whenever source is bound.
  public static void expressionConstraint(final Variable target,
    final Function function, final Variable source,
    StateRecorder recorder) throws Impossible
  {
    // Create a listener.
    Notifiable listener=new Notifiable() {
      public void boundNotify(StateRecorder irecorder) throws Impossible {
        if (isBound(source)) {
          unify(target,
            function.function(source.getValue()),irecorder);
        }
      }
    };
    source.addListener(listener,recorder);
  }
}

