package fs;

import collections.*;
import java.util.*;

public class ConceptImpl implements Concept {

  static private boolean hasChanged=true;
  static final private HashedMap conceptset = new HashedMap();
  static final private ConceptImpl TOPCONCEPT = new ConceptImpl();

  static final public ConceptImpl ANY = new ConceptImpl("Any", false);
  static final public ConceptImpl LIST = new ConceptImpl("List");
  static final public ConceptImpl ELIST = new ConceptImpl("*EList", LIST);
  static final public ConceptImpl NELIST = new ConceptImpl("*NEList", LIST,
					     new String[] {"hd","tl:List"});


  private String name = null;
  private boolean newExtensional = false, extensional = false;
  private UpdatableSet subsumes, subsumedBy;
  protected Vector directSubs = new Vector(), directSupers = new Vector();
  private UpdatableMap newApprop = new RBMap(), approp;
  private boolean restricted = true;
  Partition partition;

  ConceptImpl() {
    name = "Top";
    conceptset.putAt(name,this);
    transitiveClosure();
  }

  public ConceptImpl(String name, ConceptImpl[] directSuperList, String[] approps) {
    new Partition(this);
    if (!setName(name))
      throw new RuntimeException("Concept "+name+" already exists!");
    for (int i = 0; i < directSuperList.length; ++i)
      newIsa(directSuperList[i]);
    for (int i = 0; i < approps.length; ++i) {
      addApprop(approps[i]);
    }
    hasChanged = true;
    TOPCONCEPT.transitiveClosure();
  }

  public ConceptImpl(String name, ConceptImpl superConcept) {
    this(name, new ConceptImpl[] {superConcept}, new String[0]);
  }

  public ConceptImpl(String name) {
    this(name, TOPCONCEPT);
  }

  public ConceptImpl(String name, String[] approps) {
    this(name, new ConceptImpl[] {TOPCONCEPT}, approps);
  }

  public ConceptImpl(String name, ConceptImpl superconcept,
			String[] approps) {
    this(name, new ConceptImpl[] {superconcept}, approps);
  }

  public ConceptImpl(String name, boolean restricted) {
     this(name);
     this.restricted = restricted;
  }

  /** Return the name of this Concept. */
  public String getName() { return name; }

  /** Change the name and the extensional attribute of this Concept.
    * If the first character of the new name is a '*',
    * the concept introduces extensionality and the '*' is
    * not used for the name.
    * <p>A concept must have a unique name! If the new name already
    * exists, nothing is changed.
    */
  public boolean setName(String newname) {
    if (newname == null || newname.length() == 0)
      return true;
    boolean newextensional = false;
    if (newname.charAt(0) == '*') {
        newextensional = true;
	newname = newname.substring(1);
    }
    if (newExtensional != newextensional) {
      newExtensional = newextensional;
      hasChanged = true;
    }
    if (newname.equals(name))
      return true;
    if (isConcept(newname)) {
      System.err.println("Concept "+newname+" already exists!");
      return false;
    }
    if (name != null)
      conceptset.removeAt(name);
    name = newname;
    conceptset.putAt(name,this);
    return true;
  }

  /** Return whether this Concept is extensional. */
  public boolean isExtensional() {
    if (hasChanged)
      inheritFeatures();
    return extensional;
  }

  /** Return whether the feature <feature> is appropriate in this Concept. 
   */
  public boolean isApprop(Integer featureCode) {
    if (hasChanged)
      inheritFeatures();
    return !restricted || approp.includesKey(featureCode);
  }

  /** Return the required Type of the Value found under the given feature. */
  public Type appropType(Integer featureCode) throws NoSuchElementException {
    if (hasChanged)
      inheritFeatures();
    if (approp.includesKey(featureCode)) {
      return (Type)approp.at(featureCode);
    } else {
      if (!restricted)
         return Type.TOP;
    }
    throw new NoSuchElementException();
  }

  /** Return an Enumeration of all appropriate features. */
  public CollectionEnumeration appropFeatureCodes() {
    if (hasChanged)
      inheritFeatures();
    return approp.keys();
  }

  static public void inheritFeatures() {
    if (!hasChanged)
      return;
    hasChanged = false;
    Enumeration concepts = conceptset.keys();
    while (concepts.hasMoreElements()) {
       String conceptname = (String)concepts.nextElement();
      ConceptImpl concept = (ConceptImpl)conceptset.at(conceptname);
      concept.extensional = concept.newExtensional;
      if (concept.extensional) {
	  concept.subsumedBy = new HashedSet();
	  concept.subsumedBy.include(concept);
      } else {
	concept.subsumedBy = null;
      }
      Enumeration appropenum = concept.newApprop.keys();
      concept.approp = new RBMap();
      //System.err.println("Concept approp reset: "+concept.name);
      while (appropenum.hasMoreElements()) {
	Integer featureCode = (Integer)appropenum.nextElement();
	try {
	  concept.approp.putAt(featureCode,
			       ConjunctiveType.getType((String)concept.newApprop.at(featureCode)));
	} catch (UnificationFailure tex) {
	  System.out.println("Types for feature "+
			     Feature.getFeature(featureCode)+
			     " do not unify for concept "+concept.name+"!");
	}
      }
    }
    TOPCONCEPT.inherit();
    concepts = conceptset.keys();
    while (concepts.hasMoreElements())
      System.out.println((ConceptImpl)conceptset.at(concepts.nextElement()));      
  }

  private void inherit() {
    Enumeration subenum = directSubs.elements();
    while (subenum.hasMoreElements()) {
      ConceptImpl sub = (ConceptImpl)subenum.nextElement();
      if (extensional) {
	if (!sub.extensional) {
	  sub.extensional = true;
	  sub.subsumedBy = new HashedSet();
	  sub.subsumedBy.include(sub);
	}
	sub.subsumedBy.include(this);
      }
  // System.err.println("Concept: "+name+" approp: "+approp);
      Enumeration appropenum = approp.keys();
      while (appropenum.hasMoreElements()) {
	Integer featureCode = (Integer)appropenum.nextElement();
	Type appropType = (Type)approp.at(featureCode);
	if (sub.isApprop(featureCode)) {
	  try {
	    appropType=appropType.unify(sub.appropType(featureCode));
	  } catch (UnificationFailure uff) {
	    throw new RuntimeException("Feature "+Feature.getFeature(featureCode)+" has been refined illegally!");
	  }
	}
	sub.approp.putAt(featureCode,appropType);
      }
      sub.inherit();
    }
  }

  protected void transitiveClosure() {
    subsumes = new HashedSet();
    subsumes.include(this);
    Enumeration subenum = directSubs.elements();
    while (subenum.hasMoreElements()) {
      ConceptImpl sub = (ConceptImpl)subenum.nextElement();
      sub.transitiveClosure();
      subsumes.includeElements(sub.subsumes.elements());
    }
  }

  private void newIsa(ConceptImpl superconcept) {
    if (isa(superconcept))
      return;
    directSupers.addElement(superconcept);
    superconcept.directSubs.addElement(this);
  }

  void addIsa(ConceptImpl superconcept) throws CyclicHierarchyException {
    if (superconcept.isa(this))
      throw new CyclicHierarchyException();
    newIsa(superconcept);
    TOPCONCEPT.transitiveClosure();
  }

  void removeIsa(ConceptImpl superconcept) {
    superconcept.directSubs.removeElement(this);
    directSupers.removeElement(superconcept);
    TOPCONCEPT.transitiveClosure();
  }

  void addApprop(String approp) {
    StringTokenizer st = new StringTokenizer(approp,":");
    if (st.hasMoreElements()) {
      Integer featureCode = Feature.getCode(st.nextToken().trim());
      String typ = "";
      if (st.hasMoreElements())
	typ = st.nextToken().trim();
      newApprop.putAt(featureCode,typ);
      hasChanged = true;
    }
  }

  void removeApprops() {
    newApprop = new RBMap();
    hasChanged = true;
  }

  public void remove() {
    if (name != null)
      conceptset.removeAt(name);
    Enumeration superenum = directSupers.elements();
    while (superenum.hasMoreElements())
      ((ConceptImpl)superenum.nextElement()).directSubs.removeElement(this);
    Enumeration subenum = directSubs.elements();
    while (subenum.hasMoreElements())
      ((ConceptImpl)subenum.nextElement()).directSupers.removeElement(this);
    TOPCONCEPT.transitiveClosure();
    subsumes = null;
    directSupers = null;
    directSubs = null;
  }
 

  public static boolean isConcept(String name) {
    return conceptset.includesKey(name);
  }

  public static ConceptImpl getConcept(String name) throws TypeException {
    try {
      return (ConceptImpl)conceptset.at(name);
    } catch (NoSuchElementException e) {
      throw new TypeException();
    }
  }

  /** Return whether this Concept is-a <that> Concept.
    * In other words, return wheter this Concept is more special than <that>
    * Concept.
    */
  public boolean isa(Concept that) {
     if (that instanceof ConceptImpl) {
        return ((ConceptImpl)that).subsumes.includes(this);
     } else {
        return false;
     }
  }

  /** Return whether this Concept is-not-a <that> Concept.
    * In other words, return wheter this Concept is incompatible with <that>
    * Concept.
    */
  public boolean isNotA(Concept that) {
     if (this == that)
        return false;
     if (that instanceof ConceptImpl) {
        ConceptImpl thatCon = (ConceptImpl)that;
        if (partition == thatCon.partition)
            return true;
         return superIsNotA(thatCon) || thatCon.superIsNotA(this);
     }
     return true;
  }

  private boolean superIsNotA(ConceptImpl that) {
    Enumeration superenum = directSupers.elements();
    while (superenum.hasMoreElements()) {
       ConceptImpl superCon = (ConceptImpl)superenum.nextElement();
       if (superCon.isNotA(that))
          return true;
    }
    return false;
  }

  public CollectionEnumeration extensionalSuperconcepts() {
    if (hasChanged)
      inheritFeatures();
    return subsumedBy.elements();
  }

  public String toString() {
    if (hasChanged)
      inheritFeatures();
    StringBuffer output = new StringBuffer(name);
    if (extensional)
      output.append(" (extensional)");
    Enumeration appropenum = approp.keys();
    while (appropenum.hasMoreElements()) {
      Integer featureCode = (Integer)appropenum.nextElement();
      Type appropType = (Type)approp.at(featureCode);
      output.append("\n ").append(Feature.getFeature(featureCode))
	.append(":").append(appropType.getName());
    }
    return output.toString();
  }

}
