package fs;

import collections.*;
import java.util.*;

public class EquivRelation {

  /** Return the Unification of fs1 and fs2. */
  public static FeatureStructure unify(FeatureStructure fs1, FeatureStructure fs2)
         throws UnificationFailure {
     EquivRelation er = new EquivRelation();
     er.unify2(fs1,fs2);
     er.extensionalize();
     return er.asFeatureStructure(fs1);
  }

  /** Return the result of unifying fs2 into fs1 at the given path. */
  public static FeatureStructure unify(FeatureStructure fs1, String path,
                        FeatureStructure fs2) throws UnificationFailure {
     EquivRelation er = new EquivRelation();
     er.unify2(fs1.at(path),fs2);
     er.extensionalize();
     return er.asFeatureStructure(fs1);
  }

  /** Return whether fs1 and fs2 can be unified. */
  public static boolean canUnify(FeatureStructure fs1, FeatureStructure fs2) {
    EquivRelation er = new EquivRelation();
    try {
      er.unify2(fs1,fs2);
      return true;
    } catch (UnificationFailure uff) {
      return false;
    }
  }

  /** tie maps identity-wrapped FeatureStructures to Unificator FeatureStructures. */
  private UpdatableMap tie = new HashedMap();

  /** eit maps identity-wrapped Unificator FeatureStructures to
    * sets of identity-wrapped FeatureStructures.
    */
  private UpdatableMap eit = new HashedMap();

  /** Constructs a new Equivalence Relation. */
  public EquivRelation() {}

  /** Maps a FeatureStructure given by its Identity to a
    * Unificator FeatureStructure given by its Identity.
    * To do this, the first FeatureStructure has to be mapped to
    * the Unificator (using tie) and has to be included
    * in the set of members of the unificator (using eit).
    */
  private void map(Identity fsid, Identity uniid) {
     tie.putAt(fsid, uniid.getObject());
     UpdatableSet equivClass;
     if (eit.includesKey(uniid))
        equivClass = (UpdatableSet)eit.at(uniid);
     else {
        equivClass = new HashedSet();
        eit.putAt(uniid,equivClass);
     }
     equivClass.include(fsid);
  }

  /** Maps a FeatureStructure to a Unificator FeatureStructure.
    * If the first FeatureStructure is itself a Unificator, all its
    * elements are also mapped to the new Unificator.
    */
  private void map(FeatureStructure fs, FeatureStructure uni) {
     Identity fsid  = new Identity(fs),
              uniid = new Identity(uni);
     if (eit.includesKey(fsid)) {
       Enumeration equivClassElems=((UpdatableSet)eit.at(fsid)).elements();
       while (equivClassElems.hasMoreElements()) {
          Identity memberFSid = (Identity)equivClassElems.nextElement();
          map(memberFSid, uniid);
        }
     } else {
        map(fsid, uniid);
     }
  }


  private FeatureStructure getUnificator(FeatureStructure fs) {
     Identity fsid = new Identity(fs);
     if (tie.includesKey(fsid))
        return (FeatureStructure)tie.at(fsid);
     else
        return fs;
  }

  /** Updates the EquivRelation and constructs Unificator FeatureStructures
    * so that fs1 and fs2 are unified.
    * Other nodes may become unified by recursion.
    */
  public void unify2(FeatureStructure fs1, FeatureStructure fs2)
          throws UnificationFailure {
     Type unitype = fs1.getType().unify(fs2.getType());
     FeatureStructure uni;
     Enumeration featenum;
     Integer featureCode;
     Enumeration concepts = unitype.concepts();
     Object firstConcept;
     if (concepts.hasMoreElements() &&
         (firstConcept=concepts.nextElement()) instanceof JavaObjectFS) {
        uni = (JavaObjectFS)firstConcept;
     } else {
       UpdatableMap features = new HashedMap();
       featenum = fs1.featureCodes();
       while (featenum.hasMoreElements()) {
          featureCode = (Integer)featenum.nextElement();
          features.putAt(featureCode,fs1.delta(featureCode));
       }
       featenum = fs2.featureCodes();
       while (featenum.hasMoreElements()) {
          featureCode = (Integer)featenum.nextElement();
          features.putAt(featureCode,fs2.delta(featureCode));
       }
       uni = new FS(unitype,features);
     } /* endif */

     // Update the relation:
     // both FeatureStructures are mapped to their unificator.
     map(fs1, uni);
     map(fs2, uni);

     // search for common features of <fs1> and <fs2>:
     featenum = fs1.featureCodes();
     while (featenum.hasMoreElements()) {
        featureCode = (Integer)featenum.nextElement();
        FeatureStructure fs1post = getUnificator(fs1.delta(featureCode));
        if (fs2.hasFeature(featureCode)) {
           FeatureStructure fs2post = getUnificator(fs2.delta(featureCode));
           if (fs1post != fs2post)
             unify2(fs1post, fs2post);
        }
     }
  }

  /** Unifies all nodes in this EquivRelation according to the extensionability rule.
     */
  public void extensionalize() throws UnificationFailure {}

  /** Returns the Unificator as a FeatureStructure.
    * fs is one of the FeatureStructures that have been unified,
    * so the algorithm knows where to start.
    */
  public FeatureStructure asFeatureStructure(FeatureStructure fs) {
     Identity fsid = new Identity(fs);
     FeatureStructure uni;
     boolean expand;
     if (tie.includesKey(fsid)) {
       uni = (FeatureStructure)tie.at(fsid);
       Identity uniid = new Identity(uni);
       expand = eit.includesKey(uniid);
       if (expand)
         eit.removeAt(uniid);
     } else {
       uni = fs.newNode();
       tie.putAt(fsid, uni);
       expand = true;
     }
     if (expand) {
        Enumeration featenum = uni.featureCodes();
        while (featenum.hasMoreElements()) {
           Integer featureCode = (Integer)featenum.nextElement();
           uni.updateFeature(featureCode,asFeatureStructure(uni.delta(featureCode)));
        }
     }
     return uni;
  }

}

