package fs;

import collections.*;
import java.util.*;

public class FS extends FeatureStructure {
  Type nodetype;
  static Comparator comp = new IntComparator();
  UpdatableMap feats = new RBMap(comp);
  protected int hashCode; // as the hashCode is a finicky computation, it is being cached

  public FS(Type nodetype, Map feats) throws TypeException {
    this.nodetype = nodetype;
    hashCode = nodetype.hashCode();
    Enumeration features = feats.keys();
    while (features.hasMoreElements()) {
      Integer feature = (Integer)features.nextElement();
      addFeature(feature,(FeatureStructure)feats.at(feature));
    }
  }

  public FS(Type nodetype) {
    this.nodetype = nodetype;
    hashCode = nodetype.hashCode();
  }

  public FS() {
    this(Type.ANY);
  }

  public FS(Feature feat) throws TypeException { this(new Feature[]{feat}); }

  public FS(Type nodetype, Feature feat) throws TypeException {
    this(nodetype, new Feature[]{feat});
  }

  public FS(Feature[] features) throws TypeException {
    this(Type.TOP, features);
  }

  public FS(Type nodetype, Feature[] features) throws TypeException {
     this(nodetype);
     for (int i=0; i < features.length; ++i)
        addFeature(features[i].feature, features[i].value);
  }

  public FS(String nodetype) throws UnificationFailure {
    this(ConjunctiveType.getType(nodetype));
  }

  public FS(String nodetype, UpdatableMap feats) throws UnificationFailure, TypeException {
    this(ConjunctiveType.getType(nodetype),feats);
  }

  private FS(Type nodetype, int hashCode, UpdatableMap feats) {
    this.nodetype = nodetype;
    this.hashCode = hashCode();
    this.feats =  feats;
  }
  public Type getType() { return nodetype; }

  public CollectionEnumeration featureCodes() { return feats.keys(); }

  public boolean hasFeature(Integer featureCode) { return feats.includesKey(featureCode); }

  public FeatureStructure delta(Integer featureCode) throws NoSuchElementException {
    return (FeatureStructure)feats.at(featureCode);
  }

  public int hashCode() { return hashCode; }

  void updateFeature(Integer featureCode, FeatureStructure newValue) {
     FeatureStructure oldValue = (FeatureStructure)feats.at(featureCode);
     if (newValue != oldValue) {
       feats.putAt(featureCode, newValue);
       hashCode += featureCode.intValue() * ((newValue == this ? 1 : newValue.hashCode())
                                            -(oldValue == this ? 1 : oldValue.hashCode()));
     }
  }

  public FS addFeature(String feature, FeatureStructure value) throws TypeException {
    return addFeature(Feature.getCode(feature),value);
  }

  public FS addFeature(Integer featureCode, FeatureStructure value) throws TypeException {
     if (hasFeature(featureCode)) {
       throw new RuntimeException("Feature "+Feature.getFeature(featureCode)+" redefined!");
     }
     if (nodetype.isApprop(featureCode) && nodetype.appropType(featureCode).subsumes(value.getType())) {
       feats.putAt(featureCode, value);
       hashCode += featureCode.intValue() * (value == this ? 1 : value.hashCode());
       return this;
     } else {
        throw new TypeException();
     }
   }

  public boolean equals(Object that) {
    if (this == that)
      return true;
    if (that instanceof FeatureStructure)
      return this.subsumes((FeatureStructure)that) && ((FeatureStructure)that).subsumes(this);
    else
      return false;
  }

  public FeatureStructure newNode() {
     return new FS(nodetype, hashCode, (UpdatableMap)feats.duplicate());
  }

  public FeatureStructure without(String feature) {
    Integer featureCode = Feature.getCode(feature);
    if (!hasFeature(featureCode))
      return this;
    FS fs = (FS)newNode();
    FeatureStructure oldValue = fs.delta(featureCode);
    fs.feats.removeAt(featureCode);
    fs.hashCode -= oldValue == fs ? 1 : oldValue.hashCode();
    return fs;
  }

  private String listToString(UpdatableMap map, String indent) {
    StringBuffer output = new StringBuffer();
    if (Type.NELIST.subsumes(nodetype)) {
	  if (hasFeature(Feature.HEAD))
	      output = output.append(delta(Feature.HEAD).toString(map,indent+" "))
		.append(indent);
	  else
	    output = output.append("(no head feature!)");
	  if (hasFeature(Feature.TAIL)) {
	    FS tailfs = (FS)delta(Feature.TAIL);
	    Identity id = new Identity(this);
	    int tag = 0;
	    if (tailfs.getTag(map) != 0) {
	      output = output.append("| ")
		.append(tailfs.toString(map,indent+"   "));
	    } else {
	      output = output.append(tailfs.listToString(map,indent));
	    }
	  } else
	    output = output.append("(no tail feature!)");
    } else if (!Type.ELIST.subsumes(nodetype)) {
      output.append("| ").append(toString(map,indent+"   "));
    }
    return output.toString();
  }

  private String innerToString(UpdatableMap map, String indent) {
    StringBuffer output = new StringBuffer();
	String whitespace = "";
	if (nodetype != Type.TOP) {
	  output = output.append(nodetype.getName());
	  whitespace = indent;
	}
	CollectionEnumeration features = featureCodes();
	while (features.hasMoreElements()) {
	  Integer featureCode = (Integer)features.nextElement();
	  String feature = Feature.getFeature(featureCode);
	  output = output.append(whitespace).append(feature).append(": ")
	    .append(delta(featureCode)
		    .toString(map,indent+indent(3+feature.length())));
	  whitespace = indent;
	}
    return output.toString();
  }

  protected String untaggedToString(UpdatableMap map, String indent) {
     StringBuffer output = new StringBuffer();
     if (Type.LIST.subsumes(nodetype) && nodetype != Type.LIST) {
	  output = output.append("<").append(listToString(map,indent))
	    .append(">");
     } else {
          output.append("[").append(innerToString(map,indent)).append("]");
     }
     return output.toString();
  }

}
