package fs;

import collections.*;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;

/** The abstract class FeatureStructure defines all methods that are needed
  * to construct, access and unify Feature Structures.
  * A FeatureStructure has a Type and a number of Features that as a value
  * contain another FeatureStructure of a Type appropriate for that Feature.
  */
public abstract class FeatureStructure {


  private FeatureStructure pathFS(StringTokenizer path) {
    if (path.hasMoreTokens()) {
      String feature = (String)path.nextToken();
      if (feature.length()==0)
	return this;
      FS fs=new FS(Type.ANY);
      try {
        fs.addFeature(feature,pathFS(path));
      } catch (TypeException tee) { System.err.println("Huch?!?!?!"); }
      return fs;
    } else {
      return this;
    }
  }

  /** Construct a FeatureStructure that has this FeatureStructure as
    * the value of the given path.
    */
  public FeatureStructure pathFS(String path) {
    return pathFS(new StringTokenizer(path,":"));
  }

  /** Get the Type of the Feature Structure. */
  public abstract Type getType();

  /** Get the Type of the Feature Structure found at the given path.
    *  A path is a sequence of Features (meant to be starting at the root node).
    *  The syntax of a path is feature_1:feature_2:...feature_n.
    */
  public Type getType(String path) throws NoSuchElementException {
     return at(path).getType();
  }

  /** Test whether this Feature Structure subsumes that Feature Structure. */
  public boolean subsumes(FeatureStructure that) {
    return subsumes(new HashedMap(), that);
  }

  public boolean subsumes(UpdatableMap h, FeatureStructure that) {
    // we have to use an Identity Object as the key, because the key is
    // compared by equals() and not by == !
    Identity thisid = new Identity(this);
    if (h.includesKey(thisid))
      return h.at(thisid) == that;
    if (!getType().subsumes(that.getType()))
      return false;
    CollectionEnumeration features = this.featureCodes();
    h.putAt(thisid,that);
    while (features.hasMoreElements()) {
      Integer featureCode = (Integer)features.nextElement();
      FeatureStructure thispost = delta(featureCode);
      if (that.hasFeature(featureCode)) {
	FeatureStructure thatpost = that.delta(featureCode);
	if (!thispost.subsumes(h, thatpost))
	  return false;
      } else
	return false;
    }
    return true;
  }

  /** Test whether this Feature Structure is subsumed by that Feature Structure. */
  public boolean subsumedBy(FeatureStructure that) {
     return that.subsumes(this);
  }

  /** Calculate the unification of this Feature Structure and that Feature Structure. */
  public FeatureStructure unify(FeatureStructure that) throws UnificationFailure {
     return EquivRelation.unify(this,that);
  }

  /** Test whether this Feature Structure can be unified with that Feature Structure. */
  public boolean canUnify(FeatureStructure that) {
     return EquivRelation.canUnify(this,that);
  }

  /** Return an Enumeration of the codes of all Features of this Feature Structure. */
  public abstract CollectionEnumeration featureCodes();

  /** Returns whether this Feature Structure has the given Feature. */
  public boolean hasFeature(String feature) {
     return hasFeature(Feature.getCode(feature));
  }

  /** Returns whether this Feature Structure has the Feature given by its code. */
  public abstract boolean hasFeature(Integer featureCode);

  /** Returns the Feature Structure at the Feature given by its code. */
  public abstract FeatureStructure delta(Integer featureCode) throws NoSuchElementException;

  /** Returns the Feature Structure at the given path.
    *  A path is a sequence of Features (meant to be starting at the root node).
    *  The syntax of a path is feature_1:feature_2:...feature_n.
    *  The empty String returns the Feature Structure itself.
    */
  public FeatureStructure at(String path) throws NoSuchElementException {
    return at(new StringTokenizer(path,":"));
  }

  private FeatureStructure at(StringTokenizer path) throws
NoSuchElementException {
    if (path.hasMoreTokens())
      return (delta(Feature.getCode(path.nextToken()))).at(path);
    else
      return this;
  }

  public String toString() {
    UpdatableMap map = new HashedMap();
    buildTagMap(map,0);
    return "\n" + toString(map,"\n ") + "\n";
  }

  protected int buildTagMap(UpdatableMap tagmap, int tagcounter) {
    int tag;
    Identity id = new Identity(this);
    if (tagmap.includesKey(id)) {
      tag = ((Integer)tagmap.at(id)).intValue();
      if (tag == 0)
	tagmap.putAt(id,new Integer(++tagcounter));
    } else {
      tagmap.putAt(id,new Integer(0));
      CollectionEnumeration features = featureCodes();
      while (features.hasMoreElements()) {
	Integer featureCode = (Integer)features.nextElement();
	tagcounter = delta(featureCode).buildTagMap(tagmap, tagcounter);
      }
    }
    return tagcounter;
  }

  public final static String indent(int depth) {
    StringBuffer space = new StringBuffer(depth+1);
    for (int i = 0; i < depth; ++i)
      space.append(' ');
    return space.toString();
  }

  public final int getTag(UpdatableMap map) {
    int tag = 0;
    Identity id = new Identity(this);
    if (map.includesKey(id))
      tag = ((Integer)map.at(id)).intValue();
    return tag;
  }

  public final void setTag(UpdatableMap map, int tag) {
    map.putAt(new Identity(this),new Integer(tag));
  }

  protected String toString(UpdatableMap map, String indent) {
    StringBuffer output = new StringBuffer();
    int tag = getTag(map);
    if (tag != 0) {
      output = output.append("#").append(Math.abs(tag));
      indent = indent + indent(output.length());
    }
    if (tag >= 0) {
      if (tag != 0)
	setTag(map,-tag);
      output.append(untaggedToString(map,indent));
    }
    return output.toString();
  }

  abstract protected String untaggedToString(UpdatableMap map,
                                                            String indent);

  abstract FeatureStructure newNode();

  abstract void updateFeature(Integer featureCode, FeatureStructure
newValue);

}


